/*
 * machine.c
 * Java virtual machine interpreter.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *                  
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

/* 
 * This file has been modified by JJC, for its use as part of his debugger, bdbj.
 * Modifications are delimited by START JJC, END JJC blocks and single lines
 * modified are marked by simply JJC.
 *
 * Last modified May 2002.
 *
 * 21/5/02: Fixed bug: bad interaction between static init file access and bytecode stepping.
 * 31/5/02: Fixed bug: bad interaction between step back through exception and bytecode stepping.
 * 5/6/02:  Partially fixed bug: string utf8 failure, problem was using Cstring2Java: now happens less often.
 */

#include "debug.h"
#define	CDBG(s) 	DBG(INT_VMCALL, s)
#define	RDBG(s) 	DBG(INT_RETURN, s)
#define	NDBG(s) 	DBG(INT_NATIVE, s)
#define	IDBG(s)		DBG(INT_INSTR, s)
#define	CHDBG(s)	DBG(INT_CHECKS, s)

#include "config.h"
#include "config-std.h"
#include "config-math.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "classMethod.h"
#include "gtypes.h"
#include "bytecode.h"
#include "slots.h"
#include "icode.h"
#include "access.h"
#include "object.h"
#include "constants.h"
#include "gc.h"
#include "machine.h"
#include "lookup.h"
#include "code-analyse.h"
#include "soft.h"
#include "exception.h"
#include "external.h"
#include "baseClasses.h"
#include "thread.h"
#include "jthread.h"
#include "locks.h"
#include "checks.h"
#include "errors.h"
#include "md.h"

/* START JJC (inclusions) */
#include "code.h"   
#include <string.h> 
#include <stdlib.h> 
#include "stringSupport.h" 
#include <ctype.h>
#include "../mem/gc-mem.h"
#include "jsyscall.h"
#include "filehook.h"
#include "files.h"
/* END JJC (inclusions) */

/*
 * Define information about this engine.
 */
char* engine_name = "Interpreter";
char* engine_version = KVER;

#define	define_insn(code)	break;					\
				case code:				\
				IDBG( dprintf("%03d: %s\n", pc, #code); )
#define	define_insn_alias(code)	case code:				\
				IDBG( dprintf("%03d: %s\n", pc, #code); )

#define EXPLICIT_CHECK_NULL(_i, _s, _n)                       \
      cbranch_ref_const_ne((_s), 0, reference_label(_i, _n)); \
      softcall_nullpointer();                                 \
      set_label(_i, _n)
/* Define CREATE_NULLPOINTER_CHECKS in md.h when your machine cannot use the
 * MMU for detecting null pointer accesses */
#if defined(CREATE_NULLPOINTER_CHECKS)
#define CHECK_NULL(_i, _s, _n)                                  \
    EXPLICIT_CHECK_NULL(_i, _s, _n)
#else
#define CHECK_NULL(_i, _s, _n)
#endif

#if defined(KAFFE_PROFILER)
int profFlag;			 /* flag to control profiling */
#endif

/* START JJC (debugging macros ) */

#define JJC_DEBUG

#if defined JJC_DEBUG
#define D_FLAG(s) if( jjc_kaffe_debug ) s
#else
#define D_FLAG(s)
#endif

/* write buffers */
#define D_WB(s) D_FLAG(s)

/* read buffers */
#define D_RB(s) D_FLAG(s)

/* method calls */
#define D_FL(s) D_FLAG(s)

/* flow of simulation */
#define D_FLS(s) D_FLAG(s)

/* END JJC (debugging macros) */

/* START JJC (types and definitions) */

/* If you change these, you need to change the Java side to match */

#define MAX_BPTS 15
#define MAX_THDS 128
#define CCBRIDGE "jjc/bdbj/client/Bridge"

/* Client Calls */

#define CCgetInitialFile( ) do_execute_java_class_method( CCBRIDGE, NULL, "getInitialFile", "()Ljava/lang/String;" ).l
#define CCgetDebugFiles( )  do_execute_java_class_method( CCBRIDGE, NULL, "getDebugFiles", "()Ljava/lang/String;" ).l
#define CCgetInitialLine( ) do_execute_java_class_method( CCBRIDGE, NULL, "getInitialLine", "()I" ).i
#define CCgetBufferSize( ) do_execute_java_class_method( CCBRIDGE, NULL, "getBufferSize", "()I" ).i
#define CCgetDeepList( ) do_execute_java_class_method( CCBRIDGE, NULL, "getDeepList", "()Ljava/lang/String;" ).l
#define CCgetBreakpointLine( i ) do_execute_java_class_method( CCBRIDGE, NULL, "getBreakpointLine", "(I)I", i ).i
#define CCgetBreakpointFile( i ) do_execute_java_class_method( CCBRIDGE, NULL, "getBreakpointFile", "(I)Ljava/lang/String;", i ).l
#define CCnewThread( n, t, b ) do_execute_java_class_method( CCBRIDGE, NULL, "newThread", "(Ljava/lang/String;II)Ljava/lang/Object;", n, t, b ).l
#define CCnewMiniThread( t, b ) do_execute_java_class_method( CCBRIDGE, NULL, "newMiniThread", "(II)Ljava/lang/Object;", t ).l
#define CCthreadDead( t, b ) do_execute_java_class_method( CCBRIDGE, NULL, "threadDead", "(II)V", t, b )
#define CCminiThreadDead( t, b ) do_execute_java_class_method( CCBRIDGE, NULL, "miniThreadDead", "(II)V", t )
#define CCreportNewWatches( w ) do_execute_java_class_method( CCBRIDGE, NULL, "reportNewWatches", "(Ljava/lang/String;)V", w )
#define CCreportNewLine( l, o, t, g ) do_execute_java_class_method( CCBRIDGE, NULL, "reportNewLine", "(IIII)V", l, o, t, g )
#define CCreportNewPc( p, m, s, o, t, g ) do_execute_java_class_method( CCBRIDGE, NULL, "reportNewPc", 	\
   "(ILjava/lang/String;Ljava/lang/String;III)V", p, m, s, o, t, g )
#define CCwritePrompt( t ) do_execute_java_class_method( CCBRIDGE, NULL, "writePrompt", "(I)V", t )
#define CCcantBack( ) do_execute_java_class_method( CCBRIDGE, NULL, "cantBack", "()V" )
#define CCgetCommand( t ) do_execute_java_class_method( CCBRIDGE, NULL, "getCommand", "(I)I", t ).i
#define CCgetWatches( t ) do_execute_java_class_method( CCBRIDGE, NULL, "getWatches", "(I)Ljava/lang/String;", t ).l

#define CCgetBreakpointObject( ) do_execute_java_class_method( CCBRIDGE, NULL, "getBreakpointObject", "()Ljava/lang/Object;" ).l
#define CCaskFullThrow( o, t ) do_execute_java_class_method( CCBRIDGE, NULL, "askFullThrow", "(Ljava/lang/Object;I)V", o, t )
#define CCsafeToString( o ) do_execute_java_class_method( CCBRIDGE, NULL, "safeToString", "(Ljava/lang/Object;)Ljava/lang/String;", o ).l

int gcDisableJJC = 0;

typedef struct {
  int breakpointLine[ MAX_BPTS + 1 ];
  char breakpointFile[ MAX_BPTS + 1 ][ 512 ];
} breakpointTable;

typedef struct {
  /* CALLIN & CALLBACK */
  int hitCount;    /* if we find a certain thread keeps appearing and disappearing here, give
                    * up deallocating the debugThread buffers: its just wasting time */
  bool hasDebugAllocated;   /* is there an extant set of debugging buffers? */
  Hjava_lang_Thread* tid;
  int user_number; /* number reported to the UI */
  bool inuse;
  breakpointTable localBpts;
  /* will have to do something different here if
   * a future garbage collector started moving things around */
  Hjava_lang_Object* gcArray;
} callbackThread;
 
typedef struct {
  /* FORWARD DEBUGGING */
  int breakpoint;   /* are we running to a breakpoint? flag */
  int bytecode;     /* are we working on a bytecode scale? flag */
  int no_file_log;  /* set during certain operations to prevent logging any file changes */

  int last_obj;
  methods* last_method;

  int cbuff_push;
  int pcbuff_push;

  /* DATA LOGGING */
  int cbuffsz;

  /* circular buffer */
  slots* cbuffer_s;
  slots* cbuffer_e;
  slots* cbuffer_p;
  slots* cbuffer_v;

  /* no-push counts */
  uint16* cbuffercount_s;
  uint16* cbuffercount_e;
  uint16* cbuffercount_p;

  /* 1 means this is a object (possibly null) */
  int8* cbufferobj_s;
  int8* cbufferobj_e;
  int8* cbufferobj_p;

  bool pointer_moved;
  int32 bank;

  /* CONTROL LOGGING */
  int pcbuffsz;

  /* circular buffer for pc */
  uintp* cbufferpc_s;
  uintp* cbufferpc_e;
  uintp* cbufferpc_p;
  uintp* cbufferpc_v;

  /* Step back simulation testing: we record how far back the pointers are taken
   * by the simulation, and on a successful step back we assert that they were taken 
   * back the same distance */

  slots* cbuffer_p_test;
  slots* cbuffer_v_test;
  uintp* cbufferpc_p_test;
  uintp* cbufferpc_v_test;
  bool test_on_next_stop;
} debugThread;

/* END JJC (types) */

/* START JJC (prototypes) */

int scanstar( const char* classname, char* ilist );

int getCatch( int r );

void wcbuffs( slots value, int isobject, debugThread* debug, int8 thr );
slots rcbuffs( debugThread* debug );
bool rcbuffs_test( debugThread* debug );
void wipeOffLastInstruction( debugThread* debug );
bool rcbuffsim( debugThread* debug );
void wcbuffi( jint value, debugThread* debug );
jint rcbuffi( debugThread* debug );
void wcbuffpc( int32 value, debugThread* debug );
int32 rcbuffpc( debugThread* debug );

int fileGetBuffer( void );
void fileOpenHookJJC( int fd );
void fileOpenReverse( debugThread* debug );
void fileReadHookJJC( int fd, void* data, size_t len, ssize_t* d );
void fileReadReverse( debugThread* debug );
void fileWriteHookJJC( int fd, const void* data, size_t towrite, ssize_t* written );
void fileWriteReverse( debugThread* debug );
void fileLseekHookJJC( int fd, off_t b, int c, off_t * d );
void fileLseekReverse( debugThread* debug );
int fileCloseHookJJC( int a );

void doSetup( void );
void resetBuffers( debugThread* debug, int thr );
bool initDebug( Hjava_lang_Thread* tid, int thr );
int callbackIsFree( callbackThread* c );
int enteredNewThread( Hjava_lang_Thread* tid );
void reportThreadDeath( callbackThread* callback, debugThread* debug, int thr );
void reportNewThread( char* name, callbackThread* callback, int thr, int shouldCreate );

void restoreTopStack( volatile vmException* frame, debugThread* debug );
void askDispatch( int* currentMiniThread, bool* haveAskedDispatch, bool* anyLogged,
		  vmException* frame, vmException* lastLoggedFrame,
		  Hjava_lang_Throwable* eobj, Hjava_lang_Thread* ct );
void saveStacks( Hjava_lang_Throwable* eobj, Hjava_lang_Thread* ct );
void markException( Hjava_lang_Throwable* eobj, Hjava_lang_Thread* ct );

int isBreakpoint( uint16 line, callbackThread* callback, methods* meth, int bc );
void copyBreakpoints( callbackThread* callback );

void createLookupTable( methods* meth );
void initLocalVariables( methods* meth, int8* lvtobjs );
int lineInMethod( Method* meth, uint16 line );

void formatOperandStack( slots* opsbase, slots* opstop, int8* opsobjs, char* out, bool l );
void formatLocalVariable( char type, int index, methods* meth,
			  slots* lcl, int8* lvtobjs, char* out );
void formatLocalVariableTable( methods* meth, slots* lcl,
			       int8* lvtobjs, char* out );
void lookup_internal_value( uintp pc, methods* meth, slots* lcl, int thr, volatile vmException* mjbuf, 
			    char* in, char* out );
void getwatch( uintp pc, methods *meth, slots *lcl, callbackThread* callback, int thr, 
	       volatile vmException* mjbuf );
char get_descriptor_type( Utf8Const *descriptor );
void getToString( char *ret, Hjava_lang_Object* obj, callbackThread* callback );
void add0( char* str );
void trim( char* str );
void extractInteger( char* ret, char* str );
char* shft( char* path, char* head, char* tail,
	    uintp pc, methods* meth, slots* lcl, callbackThread* callback );
int lookup_class_field( Hjava_lang_Class *cl, char path[ 256 ], char *ret,
			uintp pc, methods* meth, slots* lcl, callbackThread* callback );
int lookup_field( Hjava_lang_Object *object, char path[ 256 ], char *ret, 
		  uintp pc, methods* meth, slots* lcl, callbackThread* callback );
int cont( int i );
int lookup_variable( uintp pc, methods *meth, slots *lcl,   
		     char path[ 256 ], char *ret, callbackThread* callback );

methods* getMethodFromPc( uintp pc, int insn, debugThread* debug,
			  bytecode* code, methods* meth );
int reverseVirtualMachine_simulation( methods* meth, debugThread* debug, callbackThread* callback, 
				      int over_mode, bool continuation, bool ignore_first,
				      volatile vmException* mjbuf );
bool doSimulation( methods* meth, debugThread* debug, callbackThread* callback,
		   int over_mode, volatile vmException* mjbuf );
int reverseVirtualMachine( methods* meth, slots* lcl, slots** spp, uintp* npcp, 
			   slots* retval, volatile vmException *mjbuf, Hjava_lang_Thread *tid,
			   debugThread* debug, callbackThread* callback, 
			   char* sourceFile, int8** opsobjsp, int8* lvtobjs,
			   int over_mode, int32* cl, bool* seen_new_line, bytecode* code,
			   int8 thr, bool ignore_first );

void generateNames( methods* meth );
int8 runVirtualMachine(methods *meth, slots *lcl, slots *sp, uintp npc, slots *retval,
		       volatile vmException *mjbuf, Hjava_lang_Thread *tid, 
		       int8 reverse_method, int8 thr );

/* END JJC (prototypes) */

/* START JJC (utility) */

int scanstar( const char* classname, char* ilist ) {
  char* c = (char*)classname;
  char* l = ilist;
  for( ; ; ) {
    c = (char*)classname;
    do {
      if( *l == '\0' )
	return *c == '\0';
    } while( *l++ == *c++ );
    l--; c--;
    if( ( *l == ' ' && *c == '\0' ) || *l == '*' )
      return 1;
    do 
      if( *l == '\0' )
	return 0;
    while( *l++ != ' ' );
  }
}

/* END JJC (utility) */

/* START JJC (extra data passing macros) */

/* The T_'s are passed TO virtualMachine/runVirtualMachine, and the F_'s are returned thereFROM.
 * the R_'s are passed to reverseVirtualMachine.   Getting the reversal of exceptions to work
 * unfortunately led to a flag explosion.    I've used up all the bits of the int8
 * except the sign bit.
 */

#define T_LOG_BIT                0x1
#define T_DEEP_BIT               0x2
#define T_BACK_BIT               0x4
#define T_OVER_BIT               0x8
#define T_RETRY_BIT             0x10
#define T_FIRST_BIT             0x20

#define T_MASK                  0x3F
#define T_CATCHOFFSET           0x40
                         
#define T_SHALLOW           0
#define T_DEEP              T_DEEP_BIT
#define T_FWDS_INTO         T_LOG_BIT 
#define T_FWDS_OVER         T_LOG_BIT + T_OVER_BIT
#define T_BACK_INTO         T_LOG_BIT + T_BACK_BIT
#define T_BACK_OVER         T_LOG_BIT + T_BACK_BIT + T_OVER_BIT
#define T_RETRY             T_LOG_BIT + T_BACK_BIT + T_RETRY_BIT 
#define T_RETRY_OVER        T_LOG_BIT + T_BACK_BIT + T_RETRY_BIT + T_OVER_BIT
#define T_RETRY_OVER_FIRST  T_LOG_BIT + T_BACK_BIT + T_RETRY_BIT + T_FIRST_BIT

#define IS_FORWARDS(r)      ( ( (r) & T_BACK_BIT  ) == 0 )
#define IS_BACKWARDS(r)     ( ( (r) & T_BACK_BIT  ) != 0 )
#define IS_RETRY(r)         ( ( (r) & T_RETRY_BIT ) != 0 )
#define IS_LOGGED(r)        ( ( (r) & T_LOG_BIT   ) != 0 )

/* possible return values of reverseVirtualMachine/runVirtualMachine
 * F_CATCH can only be returned by reverseVirtualMachine
 * and is converted to the correct RETRY mode before being
 * returned by runVirtualMachine.
 */

#define F_END              0
#define F_BEG              1
#define F_CATCH            2 
#define F_RETRY            T_RETRY
#define F_RETRY_OVER       T_RETRY_OVER
#define F_RETRY_OVER_FIRST T_RETRY_OVER_FIRST

#define R_INTO             0
#define R_OVER             1
#define R_OVERFIRST        2

int getCatch( int r ) {
  /* set catch; clear back; clear retry; clear first. */
  return ( r | T_CATCHOFFSET ) &
    ~( T_BACK_BIT & T_RETRY_BIT & T_FIRST_BIT );
}

/* END JJC (extra data passing macros) */

/* START JJC (global variables) */

int cb_set = 0;
int debugging_enabled = 0; /* three-value logic */
int pragmaJJC = 0;
callbackThread* callbackThreads[ MAX_THDS ];
debugThread* debugThreads[ MAX_THDS ]; 

uint16 debugline; /* 0 or the first line of the initital method */
char* initialfile;
char* debug_file;
char* deep_list;

uint8 defaultReverse[ 256 ];
bool jjc_kaffe_debug = false;

/* END JJC (global variables) */

int8
virtualMachine(methods*volatile meth, slots* volatile arg, slots* volatile retval, Hjava_lang_Thread* volatile tid, int8 volatile reverse, int8 volatile thr ) /* JJC (extra parameters) */
{
  methods *volatile const vmeth = meth;
  /* START JJC */
  slots *volatile const varg = arg; 
  slots *volatile const vretval = retval;
  int8 volatile const vreverse = reverse;
  bool volatile alreadyLocked = false;
  /* END JJC */

  Hjava_lang_Object* volatile mobj;
  vmException mjbuf;
  accessFlags methaccflags;

  slots* volatile lcl;
  slots* volatile sp;
  uintp volatile npc;

  int32 idx;
  jint low;

  errorInfo einfo;
  Hjava_lang_Throwable* overflow;
  jint *needOnStack;

  int8 ret = IS_BACKWARDS(reverse) ? F_BEG : F_END; /* JJC */
  int8 beg = F_END; /* JJC */

  if (tid != 0)  {
    /* implement stack overflow check */
    needOnStack = &unhand(tid)->needOnStack;

    if (jthread_stackcheck(*needOnStack) == false) {
      overflow = (Hjava_lang_Throwable*)
	unhand(tid)->stackOverflowError;
      if (overflow != 0) {
	if (*needOnStack == STACK_LOW) {
	  fprintf(stderr, 
		  "Panic: unhandled StackOverflowError()\n");
	  ABORT();
	}
	*needOnStack = STACK_LOW;
	throwException(overflow);
      }
    }
  }

  CDBG(	dprintf("Call: %s.%s%s.\n", meth->class->name->data, meth->name->data, METHOD_SIGD(meth)); )

    /* If this is native, then call the real function */
    methaccflags = meth->accflags;
  if (methaccflags & ACC_NATIVE) {

    NDBG(		dprintf("Call to native %s.%s%s.\n", meth->class->name->data, meth->name->data, METHOD_SIGD(meth)); )

      if( IS_BACKWARDS( reverse ) )   /* JJC */
	return F_BEG;               /* JJC */
    if (methaccflags & ACC_STATIC) {
      callMethodA(meth, meth, 0, (jvalue*)arg, (jvalue*)retval, 1);
    }
    else {
      callMethodA(meth, meth, ((jvalue*)arg)[0].l, &((jvalue*)arg)[1], (jvalue*)retval, 1);
    }
    return IS_BACKWARDS(reverse) ? F_BEG : F_END; /* JJC */
  }

  /* Verify method if required */
  if ((methaccflags & ACC_VERIFIED) == 0) {
    codeinfo* codeInfo;
    bool success = verifyMethod(meth, &codeInfo, &einfo);
    tidyVerifyMethod(&codeInfo);
    if (success == false) {
      throwError(&einfo);
    }
  }

  /* Allocate stack space and locals. */
  lcl = alloca(sizeof(slots) * (meth->localsz + meth->stacksz));

  mobj = 0;
  npc = 0;

  /* If we have any exception handlers we must prepare to catch them.
   * We also need to catch if we are synchronised (so we can release it).
   */
  mjbuf.logged = false; /* JJC */
  mjbuf.pc = 0;
  mjbuf.mobj = mobj;
  mjbuf.meth = meth;
  if (tid != NULL && unhand(tid)->PrivateInfo != 0) {
    mjbuf.prev = (vmException*)unhand(tid)->exceptPtr;
    unhand(tid)->exceptPtr = (struct Hkaffe_util_Ptr*)&mjbuf;
  }

  /* START JJC (so we can reverse an exception that jumps out of the logged region) */
  if( reverse == T_SHALLOW ) 
    if( JTHREAD_SETJMP( mjbuf.spbuf ) != 0 ) {
      reverse = T_RETRY;
      meth = vmeth;
      unhand(tid)->exceptPtr = (struct Hkaffe_util_Ptr*)&mjbuf;
      mobj = 0;
      npc = 0;			  			  
      mjbuf.pc = 0;
      mjbuf.mobj = mobj;
      mjbuf.meth = meth;
      unhand(tid)->exceptObj = 0;
      if( mjbuf.logged )
	thr = mjbuf.miniThread;
      else
	thr = -1;
    }

  /* END JJC */

  if (meth->exception_table != 0) {
    if (JTHREAD_SETJMP(mjbuf.jbuf) != 0) {
      meth = vmeth;
      reverse = vreverse;
      unhand(tid)->exceptPtr = (struct Hkaffe_util_Ptr*)&mjbuf;
      npc = mjbuf.pc;
      sp = &lcl[meth->localsz];
      sp->v.taddr = (void*)unhand(tid)->exceptObj;
      unhand(tid)->exceptObj = 0;
      if( mjbuf.logged )
	thr = mjbuf.miniThread;
      else
	thr = -1;
      beg = runVirtualMachine(meth, lcl, sp, npc, retval, &mjbuf, tid, getCatch( reverse ), thr ); /* JJC */
      assert( beg != F_BEG );
      /* START JJC */
      if( IS_RETRY( beg ) ) { /* we are reversing an exception */

	D_FL( fprintf( stderr, "Retry method call...\n" ) );
	reverse = beg;

	meth = vmeth;
	unhand(tid)->exceptPtr = (struct Hkaffe_util_Ptr*)&mjbuf;
	mobj = 0;
	npc = 0;			  			  
	mjbuf.pc = 0;
	mjbuf.mobj = mobj;
	mjbuf.meth = meth;
	unhand(tid)->exceptObj = 0;

	arg = varg;
	retval = vretval;
      } else {
	/* Unsync. if required */
	if (mobj != 0) {
	  unlockObject(mobj);
	}
	if (tid != NULL && unhand(tid)->PrivateInfo != 0) {
	  unhand(tid)->exceptPtr = (struct Hkaffe_util_Ptr*)mjbuf.prev;
	}
			  
	return F_END;
      }
      /* END JJC */
    }
  }

  /* START JJC */

  for( ; ; ) {
    /* Calculate number of arguments */
    idx = sizeofSigMethod(meth, false);
    if (idx == -1) {
      throwError(&einfo);
    }
    idx += (methaccflags & ACC_STATIC ? 0 : 1);
	  
    /* Copy in the arguments */
    sp = lcl;
    for (low = 0; low < idx; low++) {
      *(sp++) = *(arg++);
    }
	  
    /* Sync. if required */
    if( !alreadyLocked ) {
      alreadyLocked = true; 
      if ( methaccflags & ACC_SYNCHRONISED ) { 
	if (methaccflags & ACC_STATIC) {
	  mobj = &meth->class->head;
	}
	else {
	  mobj = (Hjava_lang_Object*)lcl[0].v.taddr;
	}
	/* this lock is safe for Thread.stop() */
	lockObject(mobj);
	/* We must store the object on which we synchronized in
	 * the mjbuf chain or else the exception handler routine
	 * won't find it.
	 */
	mjbuf.mobj = mobj;
      }       
    }
	  
    sp = &lcl[meth->localsz - 1];
	  
    ret = runVirtualMachine(meth, lcl, sp, npc, retval, &mjbuf, tid, reverse, thr ); /* JJC */
	  
    if( IS_RETRY( ret ) ) {
      meth = vmeth;
      unhand(tid)->exceptPtr = (struct Hkaffe_util_Ptr*)&mjbuf;
      mobj = 0;
      npc = 0;			  			  
      mjbuf.pc = 0;
      mjbuf.mobj = mobj;
      mjbuf.meth = meth;
      unhand(tid)->exceptObj = 0;	   
      arg = varg;
      retval = vretval;
      reverse = ret;
      continue;
    }

    /* end: JJC */
    /* Unsync. if required */
    if (mobj != 0) {
      unlockObject(mobj);
    }

    if (tid != NULL && unhand(tid)->PrivateInfo != 0) {
      unhand(tid)->exceptPtr = (struct Hkaffe_util_Ptr*)mjbuf.prev;
    }

    RDBG( dprintf( "Returning from method %s%s.\n", meth->name->data, METHOD_SIGD(meth)) ); 
    return ret; /* JJC */
  }
  /* END JJC */
}

/* START JJC (command macros) */

/* these numbers correspond to equivalent definitions in the client */

#define C_STEP          1
#define C_BREAKPOINT    2
#define C_STEPOVER      3
/* instruction 4, no longer used, was the old getwatches instruction */
#define C_STEPBACK      5
#define C_STEPBACKOVER  6
/* instruction 7, no longer used, was the old terminate instruction */
#define C_GETWATCH      8
#define C_BACKBPT       9

/* for single stepping through the bytecode */

#define C_BYTECODE_STEP         10
#define C_BYTECODE_STEPOVER     11
#define C_BYTECODE_STEPBACK     12
#define C_BYTECODE_STEPBACKOVER 13
#define C_BYTECODE_BREAKPOINT   14
#define C_BYTECODE_BACKBPT      15

/* support for a user command to reset the buffers,
 * preventing stepping back from the point at which it
 * was called, but allowing garbage collection
 */

#define C_RESET_BUFFERS         16

/* END JJC (command macros) */

/* START JJC (watch macros) */

/* type information macros - if you change these, you must
 * change the numbers in icode.h */

#define ISOBJ          1
#define NOTSUREOBJ     0
#define NOTOBJ_SKIP   -1
#define NOTOBJ_INT    -2
#define NOTOBJ_FLOAT  -3
#define NOTOBJ_LONG   -4
#define NOTOBJ_DOUBLE -5

#define GET_CONST(relevant_method, relevant_word) \
        WORD2UTF( CLASS_CONST_DATA( relevant_method->class, relevant_word ) ) 
#define W_DONE     0
#define W_UNKNOWN  1
#define W_SCOPE    2
/* value 3, no longer used, was the <object> reply: now toString is called */
#define W_LVT      4
#define W_MEMORY   5
#define W_NULL     6

/* END JJC (watch macros) */

/* START JJC (function reversal macros)  
 * special values for program counter circular stack
 * rectify() returns RECTIFY_FAIL if given START_OF_METHOD,
 * END_OF_METHOD or START_OF_BUFFER */

#define METHOD_OFFSET       0x10000

#define START_OF_METHOD     0x20000
#define END_OF_METHOD       0x20001
#define START_OF_BUFFER     0x20002 
#define START_OF_CATCH      0x20003

#define THROW_MARKER        0x40000
#define THROW_FIRST_MASK    0x00001
#define THROW_MID_MASK      0x00002
#define THROW_LAST_MASK     0x00004

#define THROW_MID         THROW_MARKER + THROW_MID_MASK  
#define THROW_FIRST       THROW_MARKER + THROW_FIRST_MASK
#define THROW_LAST        THROW_MARKER + THROW_MID_MASK   + THROW_LAST_MASK 
#define THROW_FIRST_LAST  THROW_MARKER + THROW_FIRST_MASK + THROW_LAST_MASK

/* for reversing file I/O */

#define RECTIFY_FAIL        0x80000
#define FILE_BIT           0x100000
#define FILE_OPEN          0x100001
#define FILE_CLOSE         0x100002
#define FILE_READ          0x100004
#define FILE_WRITE         0x100008
#define FILE_LSEEK         0x100010

#define NEXT_SLOT( x, s, e ) { ++(x); if( (x) == (e) ) (x) = (s); }
#define GET_NEXT_SLOT( x, s, e ) ( ( (x)+1 == (e) ) ? (s) : (x)+1 )
#define LAST_SLOT( x, s, e ) { (x) = ( ( (x) == (s) ) ? (e)-1 : (x)-1 ); }
#define GET_LAST_SLOT( x, s, e ) ( ( (x) == (s) ) ? (e)-1 : (x)-1 )

/* END JJC (function reversal macros) */

/* START JJC (buffer functions) */

#define IS_A_HEAP_POINTER(from) \
        ((uintp) (from) >= gc_heap_base && \
	 (uintp) (from) < gc_heap_base + gc_heap_range)

#if defined( JJC_DEBUG )

#define protectObject( o, t, i ) {								\
  slots array, ind, object;									\
  (array).v.taddr = callbackThreads[ t ]->gcArray;						\
  (ind).v.tint = (i);										\
  (object).v.taddr = (o);									\
  /* The next line has been removed for speed, if you are worried				\
   * that non-objects are being passed to this function you should				\
   * uncomment it */										\
  /* softcall_checkarraystore( &array, &object ); */						\
												\
  /* a quick sanity check: if claims to be an object, check that				\
     at the very least it points to somewhere in the heap					\
  */												\
  if( o != (void*)0 )										\
    if( !IS_A_HEAP_POINTER( o ) ) {								\
      fprintf( stderr, "Claimed to be in heap: %p %s:%s\n", (void*)o,				\
	 ((vmException*)unhand(getCurrentThread())->exceptPtr)->meth->class->name->data,	\
         ((vmException*)unhand(getCurrentThread())->exceptPtr)->meth->name->data );		\
      (object).v.taddr = 0;									\
    }												\
  SOFT_ADDREFERENCE( &(array), &(object) );							\
  store_offset_scaled_ref( &(array), &(ind), object_array_offset, &(object) );			\
}

#else

#define protectObject( o, t, i ) {									\
  slots array, ind, object; 										\
  (array).v.taddr = callbackThreads[ t ]->gcArray;							\
  (ind).v.tint = (i);											\
  (object).v.taddr = (o);										\
  SOFT_ADDREFERENCE( &(array), &(object) );								\
  store_offset_scaled_ref( &(array), &(ind), object_array_offset, &(object) );				\
}

#endif

#define wcbuff_increment( debug ) {		\
  ++( *( (debug)->cbuffercount_p ) );		\
  ((debug)->pointer_moved) = false;		\
}

#define wcbuff_decrement( debug ) {		\
  assert( *((debug)->cbuffercount_p) > 0 );	\
  --( *((debug)->cbuffercount_p) );		\
}

/* write a slot to the logging circular buffer */

void wcbuffs( slots value, int isobject, debugThread* debug, int8 thr ) { 
  D_WB( fprintf( stderr, "wcbuffs: value %d, obj %d, thread %d\n", value.v.tint, isobject, thr ) );       
  ++( debug->cbuff_push );

  if( debug->pointer_moved )
    *(debug->cbuffercount_p) = 0;

  *(debug->cbuffer_p) = value;
  *(debug->cbufferobj_p) = isobject;
  protectObject( ( isobject == 1 ) ? value.v.taddr : 0, (jint)thr, (jint)( (debug->cbuffer_p) - (debug->cbuffer_s) ) );
  if( debug->cbuffer_p == debug->cbuffer_v ) {
    NEXT_SLOT( debug->cbuffer_v, debug->cbuffer_s, debug->cbuffer_e );
    NEXT_SLOT( debug->cbuffer_v, debug->cbuffer_s, debug->cbuffer_e );
  }
  else if( !debug->cbuffer_v ) 
    debug->cbuffer_v = debug->cbuffer_p;
  NEXT_SLOT( debug->cbuffer_p, debug->cbuffer_s, debug->cbuffer_e );
  NEXT_SLOT( debug->cbuffercount_p, debug->cbuffercount_s, debug->cbuffercount_e );
  NEXT_SLOT( debug->cbufferobj_p, debug->cbufferobj_s, debug->cbufferobj_e );
  *( debug->cbuffercount_p ) = 0;
  (debug->pointer_moved) = true;
}

/* read a slot off the logging circular buffer */
slots rcbuffs( debugThread* debug ) {
  if( !debug->cbuffer_v ) {
    fprintf( stderr, "ABORT rcbuffs: Invalid" );
    assert( 0 );
  } else if( GET_NEXT_SLOT( debug->cbuffer_v, debug->cbuffer_s, debug->cbuffer_e ) == debug->cbuffer_p )
    debug->cbuffer_v = 0;
  LAST_SLOT( debug->cbuffer_p, debug->cbuffer_s, debug->cbuffer_e );
  LAST_SLOT( debug->cbuffercount_p, debug->cbuffercount_s, debug->cbuffercount_e );
  LAST_SLOT( debug->cbufferobj_p, debug->cbufferobj_s, debug->cbufferobj_e );
  D_RB( fprintf( stderr, "rcbuffs: value %d, object: %d, thread %d\n", (*(debug->cbuffer_p)).v.tint, 
		 *(debug->cbufferobj_p),
		 (int)debug ) ); 
  debug->last_obj = *(debug->cbufferobj_p);
  *(debug->cbufferobj_p) = 0;
  return *(debug->cbuffer_p);
}

/* try to take one value off the logging circular buffer, return true if successful */
bool rcbuffs_test( debugThread* debug ) {
  if( !debug->cbuffer_v ) {
    return false;
  } else if( GET_NEXT_SLOT( debug->cbuffer_v, debug->cbuffer_s, debug->cbuffer_e ) == debug->cbuffer_p )
    debug->cbuffer_v = 0;
  LAST_SLOT( debug->cbuffer_p, debug->cbuffer_s, debug->cbuffer_e );
  LAST_SLOT( debug->cbuffercount_p, debug->cbuffercount_s, debug->cbuffercount_e );
  LAST_SLOT( debug->cbufferobj_p, debug->cbufferobj_s, debug->cbufferobj_e );
  return true;
}

/* remove the last bytecode instructions data from the logging circular buffer */
void wipeOffLastInstruction( debugThread* debug ) {
  while( *(debug->cbuffercount_p ) == 0 )
    rcbuffs( debug );
  wcbuff_decrement( debug );
}

/* simulate the removal of a slot from the logging circular buffer,
 * without altering the counters, and keeping a note in debug->bank
 * of what the counter should be so that this function can be called
 * repeatedly giving the correct behaviour.
 */
bool rcbuffsim( debugThread* debug ) {
  if( debug->bank == -1 ) 
    debug->bank = *(debug->cbuffercount_p);
  if( debug->bank > 0 ) {
    (debug->bank)--;
    return true;
  }

  do {
    if( !debug->cbuffer_v ) {
      (debug->bank) = -1;
      return false;
    } else if( GET_NEXT_SLOT( debug->cbuffer_v, debug->cbuffer_s, debug->cbuffer_e ) == debug->cbuffer_p )
      debug->cbuffer_v = 0;
    LAST_SLOT( debug->cbuffer_p, debug->cbuffer_s, debug->cbuffer_e );
    LAST_SLOT( debug->cbuffercount_p, debug->cbuffercount_s, debug->cbuffercount_e );
    LAST_SLOT( debug->cbufferobj_p, debug->cbufferobj_s, debug->cbufferobj_e );
  } while( *(debug->cbuffercount_p) == 0 );
  (debug->bank) = *(debug->cbuffercount_p) - 1;
  return true;
}

/* push an integer to the logging circular buffer */
void wcbuffi( jint value, debugThread* debug ) {
  D_WB( fprintf( stderr, "wcbuffi: value %d, thread %d\n", value, (int)debug ) ); 

  ++( debug->cbuff_push );

  if( debug->pointer_moved )
    *(debug->cbuffercount_p) = 0;
  (*(debug->cbuffer_p)).v.tint = value;
  *(debug->cbufferobj_p) = -1; /* not an object */
  if( debug->cbuffer_p == debug->cbuffer_v ) {
    NEXT_SLOT( debug->cbuffer_v, debug->cbuffer_s, debug->cbuffer_e );
    NEXT_SLOT( debug->cbuffer_v, debug->cbuffer_s, debug->cbuffer_e );
  }
  else if( !debug->cbuffer_v ) 
    debug->cbuffer_v = debug->cbuffer_p;
  NEXT_SLOT( debug->cbuffer_p, debug->cbuffer_s, debug->cbuffer_e );
  NEXT_SLOT( debug->cbuffercount_p, debug->cbuffercount_s, debug->cbuffercount_e );
  NEXT_SLOT( debug->cbufferobj_p, debug->cbufferobj_s, debug->cbufferobj_e );
  *( debug->cbuffercount_p ) = 0;
  (debug->pointer_moved) = true;
}

/* read an integer off the logging circular buffer */
jint rcbuffi( debugThread* debug ) {
  if( !debug->cbuffer_v ) {
    fprintf( stderr, "ABORT rcbuffs: Invalid" );
    assert( 0 );
  } else if( GET_NEXT_SLOT( debug->cbuffer_v, debug->cbuffer_s, debug->cbuffer_e ) == debug->cbuffer_p )
    debug->cbuffer_v = 0;
  LAST_SLOT( debug->cbuffer_p, debug->cbuffer_s, debug->cbuffer_e );
  LAST_SLOT( debug->cbuffercount_p, debug->cbuffercount_s, debug->cbuffercount_e );
  LAST_SLOT( debug->cbufferobj_p, debug->cbufferobj_s, debug->cbufferobj_e );
  D_RB( fprintf( stderr, "rcbuffi: %d, thread %d\n", (*(debug->cbuffer_p)).v.tint, 
		 (int)debug ) ); 
  return (*(debug->cbuffer_p)).v.tint;
}

#define SPACES "                                        "

/* push a value to the program counter circular buffer */

void wcbuffpc( int32 value, debugThread* debug ) { 
  D_WB( fprintf( stderr, "%s wcbuffpc: 0x%x, thread %d\n", SPACES, value, (int)debug ) );	

  ++( debug->pcbuff_push );

  *(debug->cbufferpc_p) = value;
  if( debug->cbufferpc_p == debug->cbufferpc_v ) 
    NEXT_SLOT( debug->cbufferpc_v, debug->cbufferpc_s, debug->cbufferpc_e )
  else if( !debug->cbufferpc_v ) 
    debug->cbufferpc_v = debug->cbufferpc_p;
  NEXT_SLOT( debug->cbufferpc_p, debug->cbufferpc_s, debug->cbufferpc_e );
}

/* read a value from the program counter circular buffer */
int32 rcbuffpc( debugThread* debug ) {
  if( !debug->cbufferpc_v ) {
    fprintf( stderr, "ABORT rcbuffs: Invalid" );
    assert( 0 );
  } else if( GET_NEXT_SLOT( debug->cbufferpc_v, debug->cbufferpc_s, debug->cbufferpc_e ) == debug->cbufferpc_p )
    debug->cbufferpc_v = 0;
  LAST_SLOT( debug->cbufferpc_p, debug->cbufferpc_s, debug->cbufferpc_e );
  D_RB( fprintf( stderr, "%s rcbuffpc: 0x%x, thread %d\n", SPACES, *(debug->cbufferpc_p ), 
		 (int)debug ) );
  return *(debug->cbufferpc_p);
}

/* END JJC (buffer functions) */

/* START JJC (file I/O hooks) */

/* decide whether this file operation is going to be logged;
 * if so return the MINI-thread, if not return -1
 */

#define FILE_REVERSAL

int fileGetBuffer( void ) {

#if defined( FILE_REVERSAL )

  if( cb_set ) {
    Hjava_lang_Thread* ct = getCurrentThread( );
    vmException* frame = (vmException*)unhand(ct)->exceptPtr;
    while( frame->meth == (Method*)1 || ( frame->meth->accflags & ACC_NATIVE ) ) {
      frame = frame->prev;
      if( frame == 0 )
        return -1;
    }

    if( frame->logged == 1 ) {
      int miniThread = frame->miniThread;
      if( miniThread >= 0 && miniThread < MAX_THDS ) {
        callbackThread* callback = callbackThreads[ miniThread ];
        if( callback )
          if( callback->inuse && callback->tid == ct ) {
	    if( debugThreads[ miniThread ]->no_file_log )
	      return -1;
	    else
	      return miniThread;
          }
      }
      return -1;
    }
  }
  return -1;

#else
 
  return -1;

#endif

}

/* called on every file open */
void fileOpenHookJJC( int fd ) {
  slots s;
  debugThread* debug;
  int miniThread = fileGetBuffer( );
  if( miniThread == -1 )
    return;
  debug = debugThreads[ miniThread ];
  wcbuff_increment( debug );
  wcbuffpc( FILE_OPEN, debug );
  s.v.tint = fd;
  wcbuffs( s, -1, debug, miniThread );

  /* safety catch: this is an arbitrary number used to make us
   * more confident that we do things to the correct file
   * descriptors on reversal, if the stacks get out of synch
   * we are very unlikely to start munging random files -
   * here we really do have to be careful as these are native
   * calls doing things outside the Java `box'. */

  s.v.tint = 10103;
  wcbuffs( s, -1, debug, miniThread );

  /* all we need to reverse this is the fd, which we can only
   * get after the real call */
}

/* reversal of a logged file open */
void fileOpenReverse( debugThread* debug ) {
  assert( rcbuffs( debug ).v.tint == 10103 );
  KCLOSE_NI( rcbuffs( debug ).v.tint );
}

/* called on every file read */
void fileReadHookJJC( int fd, void* data, size_t len, ssize_t * d ) {
  off_t r;
  int rc;
  slots s;
  jlong oldpos;

  debugThread* debug;

  int miniThread = fileGetBuffer( );
  if( miniThread == -1 )
    return;
  debug = debugThreads[ miniThread ];

  rc = KLSEEK_NI( fd, 0, SEEK_CUR, &r );
  if (rc)
    return;

  wcbuff_increment( debug );
  wcbuffpc( FILE_READ, debug );
  oldpos = off_t2jlong( r );
  s.v.tlong = oldpos;
  wcbuffs( s, -1, debug, miniThread );
  s.v.tint = fd;
  wcbuffs( s, -1, debug, miniThread );
  s.v.tint = 10103;
  wcbuffs( s, -1, debug, miniThread );
}

/* reversal of logged file read */
void fileReadReverse( debugThread* debug ) {
  int fd;
  long pos;
  off_t r;
  assert( rcbuffs( debug ).v.tint == 10103 );
  fd = rcbuffs( debug ).v.tint;
  pos = rcbuffs( debug ).v.tlong;
  KLSEEK_NI( fd, jlong2off_t(pos), SEEK_SET, &r);
}

/* called on every file write */
void fileWriteHookJJC( int fd, const void* data, size_t towrite, ssize_t* written ) {
  jlong p;
  char readdata;
  slots s;
  ssize_t rr;

  int rc;
  off_t r;
  jlong len, pos;

  struct stat buf;

  int to_record;
  debugThread* debug;

  int miniThread = fileGetBuffer( );
  if( miniThread == -1 )
    return;
  debug = debugThreads[ miniThread ];

  /* get file pointer */

  rc = KLSEEK_NI( fd, 0, SEEK_CUR, &r );
  if (rc)
    return;
  pos = (off_t2jlong(r));

  /* get file length */

  rc = KFSTAT_NI( fd, &buf );
  if (rc)
    return;
  len = (off_t2jlong(buf.st_size));

  /*
    ..............................................
    |                 |
    pos               len

    =====record========
    ..............................................................
    ----towrite---------------------->| newpos

    Two cases:
    (i) pos + towrite < len:  make record of all (towrite).
    (ii) pos + towrite >= len: only record the first (len-pos).
  */

  to_record = ( ( pos + towrite < len ) ? towrite : len - pos );

  wcbuff_increment( debug );
  wcbuffpc( FILE_WRITE, debug );

  for( p = pos + to_record - 1; p != pos - 1; --p ) {
    rc = KLSEEK_NI( fd, jlong2off_t( p ), SEEK_SET, &r );
    KREAD_NI( fd, (void*)( &readdata ), jlong2off_t( 1 ), &rr );
    s.v.tint = (jint)( readdata );
    wcbuffs( s, -1, debug, miniThread );
  }

  if( to_record > 0 )
     KLSEEK_NI( fd, jlong2off_t( pos ), SEEK_SET, &r );

  s.v.tlong = to_record;
  wcbuffs( s, -1, debug, miniThread );
  s.v.tlong = pos;
  wcbuffs( s, -1, debug, miniThread );
  s.v.tint = fd;
  wcbuffs( s, -1, debug, miniThread );
  s.v.tint = 10103;
  wcbuffs( s, -1, debug, miniThread );
}

/* reversal of logged file write */
void fileWriteReverse( debugThread* debug ) {
  int fd;
  long pos, to_rewrite;

  int rc;
  off_t r;
  jlong start_pos;

  jlong i;

  unsigned char byte;
  ssize_t bwritten;

  assert( rcbuffs( debug ).v.tint == 10103 );
  fd = rcbuffs( debug ).v.tint;
  pos = rcbuffs( debug ).v.tlong;
  to_rewrite = rcbuffs( debug ).v.tlong;

  /* find out where we are to start with */

  rc = KLSEEK_NI( fd, 0, SEEK_CUR, &r );
  if (rc)
    return;
  start_pos = off_t2jlong( r );

  /* seek to beginning of block */
  rc = KLSEEK_NI( fd, jlong2off_t(pos), SEEK_SET, &r );
  if( rc )
    return;

  /* rewrite the data */

  for( i = 0; i < to_rewrite; ++i ) {
    byte = rcbuffs( debug ).v.tint;
    KWRITE_NI( fd, &byte, 1, &bwritten );
  }

  /* seek to where we started */

  rc = KLSEEK_NI( fd, jlong2off_t(pos), SEEK_SET, &r );
}

/* called on every file seek */
void fileLseekHookJJC( int fd, off_t b, int c, off_t * d ) {
  int miniThread = fileGetBuffer( );
  debugThread* debug;
  int rc;
  long oldpos;
  slots s;
  off_t r;
  if( miniThread == -1 )
    return;
  debug = debugThreads[ miniThread ];

  rc = KLSEEK_NI( fd, 0, SEEK_CUR, &r );
  if (rc)
    return;

  wcbuff_increment( debug );
  wcbuffpc( FILE_LSEEK, debug );
  oldpos = off_t2jlong( r );
  s.v.tlong = oldpos;
  wcbuffs( s, -1, debug, miniThread );
  s.v.tint = fd;
  wcbuffs( s, -1, debug, miniThread );
  s.v.tint = 10103;
  wcbuffs( s, -1, debug, miniThread );
}

/* reversal of logged file seek (same as a read) */
void fileLseekReverse( debugThread* debug ) {
  fileReadReverse( debug );
}

/* return 1 to close the file, 0 to prevent closure */
int fileCloseHookJJC( int a ) {
  int miniThread = fileGetBuffer( );
  debugThread* debug;
  if( miniThread == -1 )
    return 1;
  debug = debugThreads[ miniThread ];

  /* problem: if we want to access this file on reversal,
   * it is easiest to just keep it open since we might
   * be given a different fd if we try to open it again.
   */

  return 0;
}

void fileCloseReverse( debugThread* debug ) {
  /* noop */
}

void fileFstatHookJJC( int a, struct stat* b) {
  /* noop */
}

void fileStatHookJJC( const char* a, struct stat* b ) {
  /* noop */
}

/* END JJC (file I/O hooks) */

/* START JJC (multithreading management) */

/* ask client for various values */
void doSetup( void ) {
  int i;
  char* p;
  jstring jstr_initialfile, jstr_debug_file, jstr_deep_list;

  gcDisableJJC = (int)Kaffe_JavaVMArgs[ 0 ].gcDisableJJC;
  debugline = (int)CCgetInitialLine( );

  jstr_initialfile = (jstring)( CCgetInitialFile( ) );
  initialfile = stringJava2C( jstr_initialfile );

  jstr_debug_file = (jstring)( CCgetDebugFiles( ) );
  debug_file = stringJava2C( jstr_debug_file );
  
  jstr_deep_list = (jstring)( CCgetDeepList( ) );
  deep_list = stringJava2C( jstr_deep_list );

  /* munge the deep list */
  for( p = deep_list; *p != 0; ++p ) {
    if( *p == '.' )
      *p = '/';
    else if( *p == '$' )
      *p = '.';
  }

  for( i = 0; i < MAX_THDS; ++i )
    callbackThreads[ i ] = 0;

  for( i = 0; i < 256; ++ i )
    defaultReverse[ i ] = ( pops[ i ] < 10 ) && ( pushes[ i ] < 10 );
}

/* lookup tid in the table, if it is in the table return its
 * number, else if there is room add it to the table and return 
 * the new thread's number, else return -1 */

iLock* tidTableLockJJC;

/* set buffers back to initial state in the middle of debugging */
void resetBuffers( debugThread* debug, int thr ) {
  int j;
  debug->cbuffer_p = debug->cbuffer_s;
  debug->cbuffer_v = 0;

  debug->cbufferobj_p = debug->cbufferobj_s;

  debug->cbuffercount_p = debug->cbuffercount_s;

  debug->cbufferpc_p = debug->cbufferpc_s;
  debug->cbufferpc_v = 0;

  for( j = 0; j < debug->cbuffsz; ++j ) {
    debug->cbufferobj_s[ j ] = 0;
    debug->cbuffercount_s[ j ] = 0;
    protectObject( 0, thr, j );
  }
    
  debug->pointer_moved = false;
  debug->bank = -1;
  
  debug->no_file_log = 0;

  debug->cbuff_push = debug->pcbuff_push = 0;

  wcbuffpc( START_OF_BUFFER, debug );
  
  debug->test_on_next_stop = 0;
}

/* initiate the variables for logging a new thread,
 * return false on failure.
 */
bool initDebug( Hjava_lang_Thread* tid, int thr ) {
  int j, found;
  debugThread* debug;
  void* test;

  /* First we do a malloc to check we can create the debugThread object
   * and the garbage collection array in the client, 10k is a guess at
   * an upper limit of the size of BdbjDebugThread excluding its array. */
  int hasdebug;
  int cbuffsz, pcbuffsz;
  cbuffsz = (int)CCgetBufferSize( );
  pcbuffsz = (int)CCgetBufferSize( );

  hasdebug = callbackThreads[ thr ]->hasDebugAllocated;

  if( !hasdebug ) {
    test = gc_malloc( sizeof( debugThread ) 
		      + cbuffsz * ( sizeof( slots ) + 2 + 1 + sizeof( void* ) )
		      + pcbuffsz * ( sizeof( uintp ) ) 
		      + 10 * 1024, 
		      GC_ALLOC_FIXED );
    
    if( !test )
      return false;
    
    gc_free( test );
    debug = debugThreads[ thr ] = (debugThread*)gc_malloc( sizeof( debugThread ),
							   GC_ALLOC_FIXED );
    if( !debug )
      return false;
  } else
    debug = debugThreads[ thr ];
    
  /* At this point something horrible can happen: another thread might
   * allocate too much memory and the Java method call we will do in
   * a moment could result in an OutOfMemoryError.    For now, we do
   * everything we can to minimise the risk of this, and will fix it
   * when it becomes safe to catch OOM's in Kaffe.
   */
  
  /* if this a new-mini thread on an existent thread, or really a new thread? */
  found = -1;
  for( j = 0; j < MAX_THDS; ++j )
    if( callbackThreads[ j ] && j != thr )
      if( callbackThreads[ j ]->inuse )
	if( callbackThreads[ j ]->tid == tid ) {
	  found = j;
	  break;
	}
  if( found == -1 ) {
    callbackThreads[ thr ]->user_number = thr;
    reportNewThread( nameThread( tid ), callbackThreads[ thr ], thr, !hasdebug );
  } else {
    callbackThreads[ thr ]->user_number = callbackThreads[ j ]->user_number;
    callbackThreads[ thr ]->gcArray = CCnewMiniThread( (jint)thr, (jint)( !hasdebug ) );
  }
  
  debug->cbuffsz = cbuffsz;
  debug->pcbuffsz = pcbuffsz;

  /* allocate the buffers, one at a time: if one fails free all those allocated
   * so far and return false */

  debug->breakpoint = 0;
  
  if( !hasdebug ) {  

    debug->cbuffer_s = (slots*)gc_malloc( debug->cbuffsz * sizeof( slots ), GC_ALLOC_FIXED );
    if( !debug->cbuffer_s ) {
      gc_free( debug );
      return false;
    }

    debug->cbuffercount_s = (uint16*)gc_malloc( debug->cbuffsz * 2, GC_ALLOC_FIXED );
    if( !debug->cbuffercount_s ) {
      gc_free( debug->cbuffer_s );
      gc_free( debug );
      return false;
    }
    
    debug->cbufferobj_s = (int8*)gc_malloc( debug->cbuffsz, GC_ALLOC_FIXED );
    if( !debug->cbufferobj_s ) {
      gc_free( debug->cbuffercount_s );
      gc_free( debug->cbuffer_s );
      gc_free( debug );
      return false;
    }
  
    debug->cbufferpc_s = (uintp*)gc_malloc( debug->pcbuffsz * sizeof( uintp ), GC_ALLOC_FIXED );
    if( !debug->cbufferpc_s ) {
      gc_free( debug->cbufferobj_s );
      gc_free( debug->cbuffercount_s );
      gc_free( debug->cbuffer_s );
      gc_free( debug );
      return false;
    }

    callbackThreads[ thr ]->hasDebugAllocated = 1;
  }

  debug->cbuffer_e = debug->cbuffer_s + debug->cbuffsz - 1;
  debug->cbuffer_p = debug->cbuffer_s;
  debug->cbuffer_v = 0;

  debug->cbuffercount_e = debug->cbuffercount_s + cbuffsz - 1;
  debug->cbuffercount_p = debug->cbuffercount_s;

  debug->cbufferobj_e = debug->cbufferobj_s + cbuffsz - 1;
  debug->cbufferobj_p = debug->cbufferobj_s;

  debug->cbufferpc_e = debug->cbufferpc_s + pcbuffsz - 1;
  debug->cbufferpc_p = debug->cbufferpc_s;
  debug->cbufferpc_v = 0;

  /* reset first few items of obj and count to 0, just
   * to make sure it all goes smoothly. */

  for( j = 0; j < ( debug->cbuffsz > 6 ? 6 : debug->cbuffsz); ++j ) {
    debug->cbufferobj_s[ j ] = 0;
    debug->cbuffercount_s[ j ] = 0;
  }
    
  debug->pointer_moved = false;
  debug->bank = -1;
  
  debug->cbuff_push = debug->pcbuff_push = 0;

  wcbuffpc( START_OF_BUFFER, debug ); 
  
  debug->test_on_next_stop = 0;
  
  callbackThreads[ thr ]->inuse = true;
  callbackThreads[ thr ]->tid = tid;
  return true;
}

/* return 1 if this callback is not currently in use */
int callbackIsFree( callbackThread* c ) {
  if( !c )
    return 1;
  if( c->inuse == false )
    return 1;
  return 0;
}

/* auxillary function for enteredNewThread.
 * Call this within a tidTableLock.
 * Returns -1 if there is nowhere suitable.
 */
int findBestThreadSlot( Hjava_lang_Thread* tid ) {
  /* first look for the first free slot with the same tid:
   * this tries to get a thread that keeps flitting in and out
   * of existence fixed to a slot: if found increment
   * hit count. */
  int i;
  for( i = 0; i < MAX_THDS; ++i ) 
    if( callbackIsFree( callbackThreads[ i ] ) ) 
      if( callbackThreads[ i ] )
	if( callbackThreads[ i ]->tid == tid ) {
	  ++( callbackThreads[ i ]->hitCount );
	  return i;
	}

  /* now any free slot will do: this time reset the counter
   * to 0 as it's a different thread. */
  
  for( i = 0; i < MAX_THDS; ++i )
    if( callbackIsFree( callbackThreads[ i ] ) ) {
      if( !callbackThreads[ i ] ) {
	callbackThreads[ i ] = (callbackThread*)gc_malloc( sizeof( callbackThread ), GC_ALLOC_FIXED );
	if( !callbackThreads[ i ] )
	  return -1;
	else
	  callbackThreads[ i ]->hasDebugAllocated = 0;
      }
      ( callbackThreads[ i ]->hitCount ) = 0;
      return i;
    }

  /* oh well, this thread will just have to run unlogged then.
   */

  return -1;
}

/* called when a new mini-thread is entered, return the
 * new mini-thread number or -1 on failure
 */
int enteredNewThread( Hjava_lang_Thread* tid ) {
  int i;
  int iLockRoot;
  lockStaticMutex( &tidTableLockJJC );
  i = findBestThreadSlot( tid );
  if( i != -1 )
    if( !initDebug( tid, i ) )
      i = -1;
  unlockStaticMutex( &tidTableLockJJC );
  return i;
}

/* called when a mini-thread returns or is terminated by an exception */
void reportThreadDeath( callbackThread* callback, debugThread* debug, int thr ) {
  int iLockRoot;
  int i, found;
  int shouldDelete;
  lockStaticMutex( &tidTableLockJJC );
  for( i = 0; i < debug->cbuffsz; ++i )
    protectObject( 0, thr, i );
  callback->inuse = false;

  shouldDelete = !( callback->hitCount > 5 );

  if( shouldDelete ) {
    gc_free( debug->cbuffer_s );
    gc_free( debug->cbuffercount_s );
    gc_free( debug->cbufferobj_s );
    gc_free( debug->cbufferpc_s );
    gc_free( debug );
    callback->hasDebugAllocated = 0;
  }

  /* now does it have a brother left, if not report the real thread's death */
  found = -1;
  for( i = 0; i < MAX_THDS; ++i ) 
    if( callbackThreads[ i ] && ( i != thr ) )
      if( callbackThreads[ i ]->inuse && callbackThreads[ i ]->user_number == callback->user_number ) {
	found = i;
	break;
      }
  if( found == -1 )
    CCthreadDead( callback->user_number, shouldDelete );
  else
    CCminiThreadDead( thr, shouldDelete );
  callback->user_number = -1;

  unlockStaticMutex( &tidTableLockJJC );
}

/* report new thread to client, giving its name */
void reportNewThread( char* name, callbackThread* callback, int thr, int shouldCreate ) {
  Utf8Const* utf_name = utf8ConstNew( name, -1 );
  jstring jstr_name = utf8Const2Java( utf_name ); // stringC2Java( name );
  callback->gcArray = CCnewThread( jstr_name, (jint)( thr ), (jint)shouldCreate );
  utf8ConstRelease( utf_name );
}

/* END JJC (multithreading management) */

/* START JJC (exception management) */

/* someway in the middle of reversing an exception, restore one stack frame's
 * operand stack and local variables
 */
void restoreTopStack( volatile vmException* frame, debugThread* debug ) {
  /* the frame is new, not the one before the exception! */
  int opsdepth, i;
  /* copy in the locals */
  for( i = frame->meth->localsz - 1; i >= 0; --i ) {
    (&(frame->lclbase)[i])[ 0 ] = rcbuffs( debug );
    frame->lvtobjs[ i ] = debug->last_obj; /*   *( debug->cbufferobj_p ); */
  }
  /* have an empty operand stack, fill it */
  opsdepth = rcbuffi( debug );
  (frame->opstop) += opsdepth;
  (frame->opsobjs) += opsdepth;
  for( i = opsdepth - 1; i >= 0; --i ) {
    (&(frame->opstop)[-i])[ 0 ] = rcbuffs( debug );
    *( (frame->opsobjs) - i ) = debug->last_obj; /*  *( debug->cbufferobj_p ); */
  }
  wcbuff_decrement( debug );
} 

/* an exception has thrown out of a thread; ask the user if they
 * wish to allow the exception to dispatch or to step back to the
 * point just before it was thrown.
 */
void askDispatch( int* currentMiniThread, bool* haveAskedDispatch, bool* anyLogged,
		  vmException* frame, vmException* lastLoggedFrame,
		  Hjava_lang_Throwable* eobj, Hjava_lang_Thread* ct ) {
  int response;
  if( !*haveAskedDispatch ) {
    *haveAskedDispatch = true;
    
    /* but first, if we cannot reverse the exception just let it throw */
    debugThreads[ *currentMiniThread ]->breakpoint = false;
    if( doSimulation( lastLoggedFrame->meth, debugThreads[ *currentMiniThread ], 
		      callbackThreads[ *currentMiniThread ], R_INTO, frame ) ) { 
      CCaskFullThrow( eobj, (jint)(callbackThreads[ *currentMiniThread ]->user_number) );
      do {
	response = CCgetCommand( (jint)(callbackThreads[ *currentMiniThread ]->user_number) );
      } while( !( response == C_STEP || response == C_STEPBACK ) );
      
      if( response == C_STEPBACK ) {
	unhand(ct)->needOnStack = STACK_HIGH;
	JTHREAD_LONGJMP(lastLoggedFrame->spbuf, 1);	
	
	/* no return */
	assert( 0 );
      }	
    }
  }
  reportThreadDeath( callbackThreads[ *currentMiniThread ], debugThreads[ *currentMiniThread ], *currentMiniThread );
  if( frame->logged )
    *currentMiniThread = frame->miniThread;
  else
    *currentMiniThread = -1;
  *anyLogged = false;
}

/* an exception has occurred, run up the call stack logging the
 * stack frames and dealing with threads that are thrown out of.
 */
void saveStacks( Hjava_lang_Throwable* eobj, Hjava_lang_Thread* ct ) {
  exceptionInfo einfo;
  vmException* frame;
  vmException* lastLoggedFrame = NULL;
  int opsdepth, i;
  bool anyLogged = false;
  debugThread* debug = 0;

  bool haveAskedDispatch = false;
  int currentMiniThread = -1;

  /* New plan: call dispatch? for deepest logged mini-thread that dies.
   *           free buffers for those that die; only need to log the shallowest.
   */

  for (frame = (vmException*)unhand(ct)->exceptPtr; frame != 0; frame = frame->prev) {

    D_FL( fprintf( stderr, "-------Logging a Throw Frame-------\n" ) );
    if( ( !(frame->logged) || frame->meth == (Method*)1 || frame->miniThread != currentMiniThread ) && anyLogged ) {
      /* The exception is thrown out of the logged region, we must now ask the
       * user whether they want the exception fully thrown, or to reverse back
       * to just before the throw.   Once this exception is fully thrown there
       * is no going back. 
       */
      if( anyLogged )
	wcbuffpc( rcbuffpc( debug ) | THROW_LAST_MASK, debug );
      askDispatch( &currentMiniThread, &haveAskedDispatch, &anyLogged,
		   frame, lastLoggedFrame, eobj, ct );
    }

    /* note that we will just skip over unlogged methods backwards as usual,
     * and the shallow unlogged region (if there is one) cannot be stepped back into.
     *
     * shallow unlogged ---calls---> logged region ---calls---> deep unlogged region
     */

    if( frame->meth != (Method*)1 ) {
      if( frame->logged ) {
        currentMiniThread = frame->miniThread;
        debug = debugThreads[ currentMiniThread ];
        wcbuff_increment( debug );
        opsdepth = frame->opstop - frame->opsbase;
        for( i = 0; i < opsdepth; ++i )
	  wcbuffs( (&(frame->opstop)[-i])[ 0 ], frame->opsobjs ? *( (frame->opsobjs) - i ) : 0, debug, currentMiniThread );
        wcbuffi( opsdepth, debug );
        for( i = 0; i < frame->meth->localsz; i++ )
	  wcbuffs( (&(frame->lclbase)[i])[ 0 ], frame->opsobjs ? frame->lvtobjs[ i ] : 0, debug, currentMiniThread );
	
        wcbuffpc( frame->pc, debug );
        wcbuffpc( (uintp)(frame->meth), debug );

        if( !anyLogged ) {
	  wcbuffpc( THROW_FIRST, debug );
	  D_FL( fprintf( stderr, "THROW_FIRST\n" ) );
	  lastLoggedFrame = frame;
	  anyLogged = true;

        } else {
	  wcbuffpc( THROW_MID, debug );
	  D_FL( fprintf( stderr, "THROW_MID\n" ) );
	}
        D_FL( fprintf( stderr, "Logged frame for method: %s on miniThd %d\n", frame->meth->name->data, currentMiniThread ) );
      }
      if( findExceptionBlockInMethod(frame->pc, eobj->base.dtable->class, frame->meth, &einfo) == true ) {      
        if( anyLogged )
	  wcbuffpc( rcbuffpc( debug ) | THROW_LAST_MASK, debug );
	return;
      }
    }
  }
  askDispatch( &currentMiniThread, &haveAskedDispatch, &anyLogged,
	       frame, lastLoggedFrame, eobj, ct );  
}

/* An exception has occurred: find out on which thread and deal with it. */
void markException( Hjava_lang_Throwable* eobj, Hjava_lang_Thread* ct ) {
  int iLockRoot;
  int i;
  lockStaticMutex( &tidTableLockJJC );
  for( i = 0; i < MAX_THDS; ++i )
    if( callbackThreads[ i ] && cb_set ) {
      if( (callbackThreads[ i ])->inuse )
        if( ct == callbackThreads[ i ]->tid ) {
          unlockStaticMutex( &tidTableLockJJC );
          if( debugThreads[ i ] ) {
            saveStacks( eobj, ct );
	  }
	  return;
	}
    }
  unlockStaticMutex( &tidTableLockJJC );
}

/* END JJC (exception management) */

/* START JJC (breakpoint management) */

#define IS_BREAKPOINT( line, bc )											\
  do {															\
    is_breakpoint = 0;													\
    for( i = 0; i < MAX_BPTS; ++i ) {											\
      if( callback->localBpts.breakpointLine[ i ] == -1 ) {								\
	is_breakpoint = 0;												\
	break;														\
      }															\
      else {														\
	if( callback->localBpts.breakpointLine[ i ] == (line) )								\
	  if( strcmp( callback->localBpts.breakpointFile[ i ],								\
		      (bc) ? meth->bytecodeBreakpointSig : meth->fullSourceFile ) == 0 ) {				\
	    is_breakpoint = 1;												\
	    break;													\
	  }														\
      }															\
    }															\
  } while( 0 )										
   

/* return true if the given line/pc is a breakpoint (bc=true means use pc/class
 * rather than line/file.
 */
int isBreakpoint( uint16 line, callbackThread* callback, methods* meth, int bc ) {
  int is_breakpoint, i;
  IS_BREAKPOINT( line, bc );
  return is_breakpoint;
}

/* get a local copy of the current breakpoint table from the client */
void copyBreakpoints( callbackThread* callback ) {
  int i;
  int iLockRoot;
  Hjava_lang_Object* breakpointObject;
  breakpointObject = (Hjava_lang_Object*)CCgetBreakpointObject( );
  lockMutex( breakpointObject );
  for( i = 0; i < MAX_BPTS; ++i ) {
    jstring file;
    callback->localBpts.breakpointLine[ i ] = (int)CCgetBreakpointLine( i );
    file = (jstring)( CCgetBreakpointFile( i ) );
    stringJava2CBuf( file, callback->localBpts.breakpointFile[ i ], 511 );
  }
  unlockMutex( breakpointObject );
  callback->localBpts.breakpointLine[ MAX_BPTS ] = -1;
}

/* END JJC (breakpoint management) */

/* START JJC (program counter management) */

#define rectify( pc ) ( (pc) & FILE_BIT || (pc) & START_OF_METHOD || (pc) & THROW_MARKER ? RECTIFY_FAIL : ( (pc) & 0xFFFF ) )

/* for speed, generate a pc->line number lookup table for this method */
void createLookupTable( methods* meth ) {
  int32 i, j;
  uint16 pc, line, size, nentries;

  /* sort the line number table 
   * we use insertion sort because:
   * (i) the table is generally quite small,
   * (ii) almost all compilers produce a sorted table
   * but we should do a sort, just in case.
   */
  nentries = meth->lines->length;
  for( j = 1; j < nentries; ++j ) {
    pc = meth->lines->entry[ j ].start_pc;
    line = meth->lines->entry[ j ].line_nr;
    i = j - 1;
    while( ( i >= 0 ) && ( meth->lines->entry[ i ].start_pc > pc ) ) {
      meth->lines->entry[ i + 1 ] = meth->lines->entry[ i ];
      --i;
    }
    meth->lines->entry[ i + 1 ].start_pc = pc;
    meth->lines->entry[ i + 1 ].line_nr = line;
  }

  /* create line number lookup table */
  size = meth->c.bcode.codelen;
  nentries = meth->lines->length;
  meth->lookup = (uint16*)gc_malloc( 2 * size, GC_ALLOC_FIXED );
  if( !meth->lookup ) {
    fprintf( stderr, "Out of Memory\n" );
    assert( 0 );
  }

  /* the while loop is intended to deal with the instance where we have two
   * line numbers mapping onto the same pc */
  
  for( i = 0, j = 0, line = meth->lines->entry[ 0 ].line_nr; i < size; ++i ) {
    if( ( j != nentries - 1 ) && ( i == meth->lines->entry[ j + 1 ].start_pc ) ) {
      do {
        ++j;
        line = meth->lines->entry[ j ].line_nr;
      } while( ( j != nentries - 1 ) && ( meth->lines->entry[ j ].start_pc == 
					  meth->lines->entry[ j + 1 ].start_pc ) );
    }
    meth->lookup[ i ] = line;
  } 
}

/* work out types of arguments, set others to type currently unknown. */
void initLocalVariables( methods* meth, int8* lvtobjs ) {
  int32 i, j;
  uint16 size, nentries;

  size = meth->localsz;
  for( i = 0; i < size; ++i )
    lvtobjs[ i ] = 0;

  if( (meth->accflags) & ACC_STATIC )
    j = 0;
  else {
    lvtobjs[ 0 ] = 1;
    j = 1;
  }

  nentries = meth->parsed_sig->nargs;
 
  for( i = 0; i < nentries; ++i ) {
    const char* c = ( ( meth->parsed_sig->signature->data ) + ( meth->parsed_sig->ret_and_args[ i + 1 ] ) );
    switch( *c ) {
    case '[': case 'L':
      if( strncmp( c, "Lkaffe/util/Ptr;", 16 ) == 0 )
	lvtobjs[ j++ ] = NOTSUREOBJ;
      else
	lvtobjs[ j++ ] = ISOBJ;
      break;
    case 'D':
      lvtobjs[ j++ ] = NOTOBJ_DOUBLE;
      lvtobjs[ j++ ] = NOTOBJ_SKIP;
      break;
    case 'J':
      lvtobjs[ j++ ] = NOTOBJ_LONG;
      lvtobjs[ j++ ] = NOTOBJ_SKIP;
      break;
    case 'I': case 'Z': case 'S': case 'B': case 'C':
      lvtobjs[ j++ ] = NOTOBJ_INT;
      break;
    case 'F':
      lvtobjs[ j++ ] = NOTOBJ_FLOAT;
      break;
    }
  }
  meth->aboveargs = j;

  for( i = meth->aboveargs; i < meth->localsz; ++i )
    lvtobjs[ i ] = 0;
}

#define currentLine( pc, meth ) ( ((pc)==START_OF_METHOD || (pc)==START_OF_CATCH || (pc)==RECTIFY_FAIL)?0:(meth)->lookup[pc] )

#define currentPc( pc ) ( ((pc)==START_OF_METHOD || (pc)==START_OF_CATCH || (pc)==RECTIFY_FAIL)? RECTIFY_FAIL : (pc) )

#define currentIsPc( pc, meth ) ( debug->bytecode || !(meth)->haveSource )

#define currentPoint( pc, meth ) ( ( debug->bytecode || !(meth)->haveSource ) ? \
  currentPc( rectify( pc ) ) : currentLine( rectify( pc ), (meth) ) )

/* return true if this line occurs anywhere in this method's line number table */
int lineInMethod( Method* meth, uint16 line ) {
  int32 j, nentries;
  nentries = meth->lines->length;
  for( j = 0; j < nentries; ++j ) 
    if( meth->lines->entry[ j ].line_nr == line )
      return 1;
  return 0;
}

/* END JJC (program counter management) */

/* START JJC (stack macros) */
 
#define savestack(n) { 								\
   for( i = 0; i < n; ++i )							\
     wcbuffs( stack( i )[ 0 ], opsobjs ? *( opsobjs - i ) : 0, debug, thr );	\
}

#define savestackandn(n) {			\
  savestack( n );				\
  wcbuffi( n, debug );				\
}

#define loadstack(n) {						\
   for( i = n - 1; i >= 0; i-- ) {				\
     stack( i )[ 0 ] = rcbuffs( debug );			\
     if( opsobjs )						\
        *( opsobjs - i ) = debug->last_obj;			\
        /* *( opsobjs - i ) = *( debug->cbufferobj_p ); */	\
   }								\
} 

#define loadnandstack {				\
  n = rcbuffi( debug );         \
  push( n );					\
  loadstack( n );				\
}

#define restorestack(i) {         \
  pop( pushes[ i ] - pops[ i ] ); \
  loadstack( pops[ i ] );         \
}

#define savewholestack savestackandn( sp - osp );  

#define savelocals {								\
  for( i = 0; i < meth->localsz; i++ ) {					\
     wcbuffs( local( i )[ 0 ], opsobjs ? lvtobjs[ i ] : 0, debug, thr );	\
  }										\
}

#define loadlocals {							\
   for( i = ( meth->localsz ) - 1; i >= 0; i-- ) {			\
     local( i )[ 0 ] = rcbuffs( debug );				\
     if( opsobjs )							\
       lvtobjs[ i ] = debug->last_obj; /*  *( debug->cbufferobj_p ); */	\
   }									\
}

/* END JJC (stack macros) */

/* START JJC (watches) */

#define getCharRep( c ) ( c < 0x100 && isprint( (char)c ) ? "'%c'" : "0x%x" )

/* generate a string representation of the operand stack */
void formatOperandStack( slots* opsbase, slots* opstop, int8* opsobjs, char* out, bool l ) {
  char nextHex[ 20 ];
  slots* next;
  int i;
  int type = 0;
  strcpy( out, "" );
  for( next = opstop, i = 0; next > opsbase; --next, ++i ) {
    if( opsobjs ) 
      type = *( opsobjs - i );
    switch( type ) {
    case NOTSUREOBJ: {
      if( l ) 
	sprintf( nextHex, "(%08x)%08x ", (uint32)(next[0].v.tlong >> 32),
		 (uint32)(next[0].v.tint) );
      else
	sprintf( nextHex, "%08x ", (uint32)(next[0].v.tint) );
      break;
    }
    case ISOBJ:
      sprintf( nextHex, "[%p] ", next[0].v.taddr );
      break;
    case NOTOBJ_SKIP:
      strcpy( nextHex, " " );
      break;
    case NOTOBJ_INT:
      sprintf( nextHex, "%d ", next[0].v.tint );
      break;
    case NOTOBJ_LONG:
      sprintf( nextHex, "%lldL ", next[0].v.tlong );
      break;
    case NOTOBJ_FLOAT:
      sprintf( nextHex, "%gF ", next[0].v.tfloat );
      break;
    case NOTOBJ_DOUBLE:
      sprintf( nextHex, "%gD ", next[0].v.tdouble );
      break;
    }
    if( strlen( out ) + strlen( nextHex ) > 1000 ) {
      strcat( out, "..." );
      return;
    }
    strcat( out, nextHex );
  }
  if( strcmp( out, "" ) == 0 )
    strcpy( out, "empty" );
}

/* convert a single local variable to a string representation */
void formatLocalVariable( char type, int index, methods* meth,
                          slots* lcl, int8* lvtobjs, char* out ) {
  if( index < 0 || index >= meth->localsz ) 
    sprintf( out, "<bad index>" );
  else {
    if( type == 'a' && lvtobjs ) {
      int auto_type = lvtobjs[ index ];
      type = ( auto_type == NOTOBJ_SKIP   ? '!' :
               auto_type == NOTOBJ_INT    ? 'i' :
               auto_type == NOTOBJ_LONG   ? 'l' :
               auto_type == NOTOBJ_FLOAT  ? 'f' :
               auto_type == NOTOBJ_DOUBLE ? 'd' :
               auto_type == ISOBJ         ? 'o' :
	       'x' );
    }        
    switch( type ) {
    case( '!' ): sprintf( out, "<unused> " ); break;
    case( 'i' ): sprintf( out, "%d ", (lcl[index]).v.tint ); break;
    case( 'f' ): sprintf( out, "%gF ", (lcl[index]).v.tfloat ); break;
    case( 'd' ): sprintf( out, "%gD ", (lcl[index]).v.tdouble ); break;
    case( 'l' ): sprintf( out, "%lldL ", (lcl[index]).v.tlong ); break;
    case( 'c' ): sprintf( out, getCharRep( (lcl[index]).v.tint ), 
			  (lcl[index]).v.tint ); break;
    case( 'o' ): sprintf( out, "[%p] ", (lcl[index]).v.taddr ); break;
    case( 'x' ): sprintf( out, "0x%x ", (int)(lcl[index]).v.taddr ); break;
    default: sprintf( out, "<types: i(nt) l(ong) f(loat) d(ouble) c(har) x(hex) o(bject) a(auto)>" ); break;
    }
  }
}

/* generate a string representation of the local variables */
void formatLocalVariableTable( methods* meth, slots* lcl,
                               int8* lvtobjs, char* out ) {
  char nextHex[ 30 ];
  int i;
  strcpy( out, "" );
  for( i = 0; i < meth->localsz; ++i ) {
    sprintf( nextHex, "%d: ", i );
    formatLocalVariable( 'a', i, meth, lcl, lvtobjs, nextHex );
    if( strcmp( nextHex, "<unused> " ) == 0 )
      strcpy( nextHex, " " );
    if( strlen( out ) + strlen( nextHex ) > 1000 ) {
      strcat( out, "..." );
      return;
    }
    strcat( out, nextHex );
  }
  if( strcmp( out, "" ) == 0 )
    strcpy( out, "no locals" );
}


/* special watch variables (all begin with an !) */
void lookup_internal_value( uintp pc, methods* meth, slots* lcl, int thr, volatile vmException* mjbuf, 
			    char* in, char* out ) {
  if( strcmp( in, "pc" ) == 0 )
    sprintf( out, "%d", pc );
  else if( strcmp( in, "ops" ) == 0 )
    formatOperandStack( mjbuf->opsbase, mjbuf->opstop, mjbuf->opsobjs, out, false );
  else if( strcmp( in, "opsl" ) == 0 )
    formatOperandStack( mjbuf->opsbase, mjbuf->opstop, mjbuf->opsobjs, out, true );
  else if( strcmp( in, "lbuffpush" ) == 0 )
    sprintf( out, "%d", debugThreads[ thr ]->cbuff_push );
  else if( strcmp( in, "pcbuffpush" ) == 0 )
    sprintf( out, "%d", debugThreads[ thr ]->pcbuff_push );
  else if( strcmp( in, "lbuffp" ) == 0 )
    sprintf( out, "%d", debugThreads[ thr ]->cbuffer_p - debugThreads[ thr ]->cbuffer_s );
  else if( strcmp( in, "lbuffv" ) == 0 ) {
    if( debugThreads[ thr ]->cbuffer_v ) 
      sprintf( out, "%d", debugThreads[ thr ]->cbuffer_v - debugThreads[ thr ]->cbuffer_s );
    else
      sprintf( out, "none valid" );
  }
  else if( strcmp( in, "pcbuffp" ) == 0 )
    sprintf( out, "%d", debugThreads[ thr ]->cbufferpc_p - debugThreads[ thr ]->cbufferpc_s );
  else if( strcmp( in, "pcbuffv" ) == 0 ) {
    if( debugThreads[ thr ]->cbufferpc_v )
      sprintf( out, "%d", debugThreads[ thr ]->cbufferpc_v - debugThreads[ thr ]->cbufferpc_s );
    else
      sprintf( out, "none valid" );
  }
  else if( strcmp( in, "sp" ) == 0 )
    sprintf( out, "%d", mjbuf->opstop - mjbuf->opsbase );
  else if( strcmp( in, "line" ) == 0 )
    sprintf( out, "%d", currentLine( pc, meth ) );
  else if( strcmp( in, "locals" ) == 0 )
    sprintf( out, "%d", meth->localsz );
  else if( strncmp( in, "local ", 6 ) == 0 )
    formatLocalVariable( 'a', atoi( in + 5 ), meth, lcl, mjbuf->lvtobjs, out );
  else if( strncmp( in, "local", 5 ) == 0 ) 
    formatLocalVariable( *( in + 5 ), atoi( in + 6 ), meth, lcl, mjbuf->lvtobjs, out );
  else if( strcmp( in, "lvt" ) == 0 )
    formatLocalVariableTable( meth, lcl, mjbuf->lvtobjs, out );
  else if( strcmp( in, "debug" ) == 0 ) {
    strcpy( out, "<done>" );
    jjc_kaffe_debug = true;
  }
  else if( strcmp( in, "nodebug" ) == 0 ) {
    strcpy( out, "<done>" );
    jjc_kaffe_debug = false;
  }
  else
    sprintf( out, "<bad internal variable>" );
}

/* ask the client for the watch variable that is to be looked up */
void getwatch( uintp pc, methods *meth, slots *lcl, callbackThread* callback, int thr, 
	       volatile vmException* mjbuf ) {
  jstring jstr, jstr_ret;
  char ret[ 1024 ];
  char result[ 1024 ];
  char str[ 257 ];
  char *resultp = result;
  int i;
  Utf8Const* utf_name;
  jstring jstr_name;
  jstr = (jstring)(CCgetWatches( (jint)(callback->user_number) ) );
  stringJava2CBuf( jstr, str, 256 );
  
  if( strlen( str ) > 254 )
    strcpy( result, "<identifier too long>" );
  else if( str[ 0 ] == '!' ) 
    lookup_internal_value( pc, meth, lcl, thr, mjbuf, str + 1, resultp );
  else {
    switch( lookup_variable( pc, meth, lcl, str, resultp, callback ) ) {
    case W_UNKNOWN: strcpy( result, "<unknown>" ); break;
    case W_SCOPE  : strcpy( result, "<not in scope>" ); break;
    case W_LVT    : strcpy( result, "<no local variable table>" ); break;
    case W_MEMORY : strcpy( result, "<out of memory>" ); break;
    case W_NULL   : strcpy( result, "null" ); break;
    }
  }
  /* change any new line \n in nextvalue to a \ */
  for( i = 0; i < strlen( result ); i++ ) 
    if( result[ i ] == '\n' )
      result[ i ] = '\\';
  
  if( strlen( str ) + strlen( result ) + 3 < 1023 ) {
    strcpy( ret, str );
    strcat( ret, " == " );
    strcat( ret, result );
  }
  
  /* Send the results back to Java */  

  utf_name = utf8ConstNew( ret, -1 );
  jstr_name = utf8Const2Java( utf_name );
  jstr_ret = utf8Const2Java( utf_name );
  CCreportNewWatches( jstr_ret );
  utf8ConstRelease( utf_name );
}

/* Returns 
           other = unmatched, i.e. object or array
           T = string
   B, C, D, F, I, J, S, Z 
*/
char get_descriptor_type( Utf8Const *descriptor ) {
  char first =  *( descriptor->data );
  if( first == 'L' || first == '[' ) {
    Utf8Const* str = utf8ConstNew( "Ljava/lang/String;", -1 );
    if( utf8ConstEqual( str, descriptor ) ) {
      utf8ConstRelease( str );
      return ('T');
    }
    utf8ConstRelease( str );
    return ('U');
  } else
    return first;
}

/* Function to call the toString() method of an object
 * This method always exists, as it's a method of Object.
 * The client method which calls toString catches any exception.
 */
void getToString( char *ret, Hjava_lang_Object* obj, callbackThread* callback ) {
  jstring jstr;
  char str[ 1027 ];
  jstr = (jstring)( CCsafeToString( obj ) );
  stringJava2CBuf( jstr, str, 1026 );
  if( strlen( str ) + 4 < 1024 ) {
    strcpy( ret, str );
  } else
    strcpy( ret, "<string too long>" );
}

/* For floats and doubles, if there is decimal point
 * add .0, otherwise integral floats could be used to index arrays.
 */
void add0( char* str ) {
  if( !strchr( str, '.' ) && !strchr( str, 'e' )
      && !strchr( str, 'E' ) )
    strcat( str, ".0" );
}

/* Trim spaces from a string at the beginning and end. */
void trim( char* str ) {
  char* end;
  char* start;
  end = str + strlen( str ) - 1;
  while( *end == ' ' )
    end--;
  *( end + 1 ) = '\0';
  start = str;
  while( *start == ' ' ) 
    start++;
  do 
    *str++ = *start++;
  while( *( start - 1 ) != '\0' );
}

/* Convert the value of the index computed into an integer
 * string if possible (in many cases there is nothing to do).
 * Return a ">" if not possible (this will ensure that
 * the lookup fails and returns <bad index> when its time comes.
 */
void extractInteger( char* ret, char* str ) {
  int isNumber = 1;
  char* ptr;
  trim( str );
  if( strchr( "<\"tf", *str ) ) {
    strcpy( ret, ">" );
    return;
  }
  if( *str == '\'' ) {
    int val = (int)(*( str + 1 ));
    sprintf( ret, "%d", val );
    return;
  }
  if( *str == '0' && *( str + 1 ) == 'x' ) {
    int val;
    sscanf( str, "%x", &val );
    sprintf( ret, "%d", val );
    return;
  }
  for( ptr = str; *ptr != '\0'; ptr++ )
    if( !isdigit( *ptr ) )
      isNumber = 0;
  if( !isNumber )
    strcpy( ret, ">" );
  else
    strcpy( ret, str );
}

/* Separate Head From Tail, expanding variable array subscripts 
 * a[b][c[d]].e[f] becomes head = a, tail = b'.c'.e[f]
 * where b evaluates to b' and c[d] evaluates to c'.
 * If it is not an integer put the character ">" in instead,
 * which should cause a suitable error message later.
 */
char* shft( char* path, char* head, char* tail,
	    uintp pc, methods* meth, slots* lcl, callbackThread* callback ) {
  char* bpos;
  char* dpos = strchr( path, '.' );
  char* ind;
  char next_index[ 12 ] = "";
  char temp_tail[ 256 ] = "";
  char temp_value[ 1024 ] = "";

  if( !dpos )
    strcpy( head, path );
  else {
    strcpy( tail, dpos + 1 );
    strncpy( head, path, dpos - path );
    head[ dpos - path ] = '\0';
  }

  trim( head );
  trim( tail );

  bpos = strchr( head, '[' );
  if( !bpos ) 
    return dpos;

  ind = bpos;
  while( *ind == '[' ) {
    int count = 1;
    char* location;
    for( location = ind + 1;
	 count != 0;
	 *location == '[' ? ++count : *location == ']' ? --count : 0, ++location )
      ;
    *( location - 1 ) = '\0';
    extractInteger( temp_value, ind + 1 );
    if( *temp_value == '>' )
      lookup_variable( pc, meth, lcl, ind + 1, temp_value, callback );
    extractInteger( next_index, temp_value );      
    if( strlen( temp_tail ) + strlen( next_index ) + 1 < 255 ) {
      strcat( temp_tail, next_index );
      strcat( temp_tail, "." );
    }
    ind = location;
    while( *ind == ' ' )
      ind++;
  }
  if( strlen( temp_tail ) + strlen( tail ) < 255 ) {
    strcat( temp_tail, tail );
  }
  strcpy( tail, temp_tail );
  if( tail[ strlen( tail ) - 1 ] == '.' )
    tail[ strlen( tail ) - 1 ] = '\0';
  head[ bpos - head ] = '\0'; 

  trim( head );
  trim( tail );
  return bpos;  
}

/* Lookup a static field, return is one of W_DONE, ... */
int lookup_class_field( Hjava_lang_Class *cl, char path[ 256 ], char *ret,
			uintp pc, methods* meth, slots* lcl, callbackThread* callback ) {
  Utf8Const *utf8;
  int j, nsf;
  Field *fld;
  slots s;

  /* Separate head and tail */
  char *pos;
  char head[ 256 ] = "";
  char tail[ 256 ] = "";
  pos = shft( path, head, tail,
	      pc, meth, lcl, callback );

  /* check that the class has been loaded */

  if( cl->state != CSTATE_COMPLETE ) {
    strcpy( ret, "<class loading not complete>" );
    return W_DONE;
  }

  /* now create a utf8 for the head and "java/lang/String" */
  utf8 = utf8ConstNew( head, -1 );
  if( !utf8 )
    return W_MEMORY;

  nsf = cl->nsfields;
  for( j = 0; j < nsf; j++ ) {
    if( utf8ConstEqual( ( &( cl->fields )[ j ] )->name, utf8 ) ) {
      fld = &( cl->fields )[ j ];
      if( FIELD_ISREF( fld ) && !pos ) {
        /* test for String */
        errorInfo info;
        Hjava_lang_Class *stringclass = lookupClass( "java/lang/String", NULL, &info );
        if( fld->type == stringclass ) {
          jstring jstr;
          char str[ 1027 ];
          load_addr_ref( &s, FIELD_ADDRESS( fld ) );
          jstr = (jstring)( s.v.taddr );
          stringJava2CBuf( jstr, str, 1026 );
          if( strlen( str ) + 2 < 1024 ) {
            strcpy( ret, "\"" );
            strcat( ret, str );
            strcat( ret, "\"" );
          } else
            strcpy( ret, "<string too long>" );
          utf8ConstRelease( utf8 );
          return W_DONE;
        } else {
          utf8ConstRelease( utf8 );
	  load_addr_ref( &s, FIELD_ADDRESS( fld ) );
	  if( (void *)( s.v.taddr ) == 0 )
	    return W_NULL;
	  else {
	    getToString( ret, s.v.taddr,
			 callback );
	    return W_DONE;
	  }
        }
      } else if( FIELD_ISPRIM( fld ) && !pos ) {
	int c;
	switch( CLASS_PRIM_SIG( fld->type ) ) {
	case 'B':
	  load_addr_byte( &s, FIELD_ADDRESS( fld ) );
	  sprintf( ret, "%d", s.v.tint );
	  utf8ConstRelease( utf8 );
	  return W_DONE;
	case 'C':
	  load_addr_char( &s, FIELD_ADDRESS( fld ) );
	  c = s.v.tint;
	  if( ( c < 0x100 ) && isprint( (char)c ) )
	    sprintf( ret, "'%c'", (char)c );
	  else
	    sprintf( ret, "0x%x", c );
	  utf8ConstRelease( utf8 );
	  return W_DONE;
	case 'D':
	  load_addr_double( &s, FIELD_ADDRESS( fld ) );
	  sprintf( ret, "%g", s.v.tdouble );
	  add0( ret );
	  utf8ConstRelease( utf8 );
	  return W_DONE;
	case 'F':
	  load_addr_float( &s, FIELD_ADDRESS( fld ) );
	  sprintf( ret, "%g", s.v.tfloat );
	  add0( ret );
	  utf8ConstRelease( utf8 );
	  return W_DONE;
	case 'I':
	  load_addr_int( &s, FIELD_ADDRESS( fld ) );
	  sprintf( ret, "%d", s.v.tint );
	  utf8ConstRelease( utf8 );
	  return W_DONE;
	case 'J':
	  load_addr_long( &s, FIELD_ADDRESS( fld ) );
	  sprintf( ret, "%lld", s.v.tlong );
	  utf8ConstRelease( utf8 );
	  return W_DONE;
	case 'S':
	  load_addr_short( &s, FIELD_ADDRESS( fld ) );
	  sprintf( ret, "%d", s.v.tint );
	  utf8ConstRelease( utf8 );
	  return W_DONE;
	case 'Z':
	  load_addr_byte( &s, FIELD_ADDRESS( fld ) );
	  if( s.v.tint == 0 )
	    strcpy( ret, "false" );
	  else
	    strcpy( ret, "true" );
	  utf8ConstRelease( utf8 );
	  return W_DONE;
	}
      } else if( FIELD_ISREF( fld ) ) {
        load_addr_ref( &s, FIELD_ADDRESS( fld ) );
        utf8ConstRelease( utf8 );
        return lookup_field( (Hjava_lang_Object *)( s.v.taddr ), tail, ret, 
			     pc, meth, lcl, callback );
      }
    }
  }
  utf8ConstRelease( utf8 );
  return W_UNKNOWN;
}

#define FIELD_IS_STATIC(j) (j < nsf) 

/* Lookup a non-static field, return is one of W_DONE, ... */
int lookup_field( Hjava_lang_Object *object, char path[ 256 ], char *ret, 
		  uintp pc, methods* meth, slots* lcl, callbackThread* callback ) {
  Hjava_lang_Class *cl;
  Utf8Const *utf8;
  int j, nf, nsf, ch;
  Field *fld;
  slots s;

  /* Separate head and tail */
  char *pos;
  char head[ 256 ] = "";
  char tail[ 256 ] = "";
  pos = shft( path, head, tail,
	      pc, meth, lcl, callback );

  if( (void *)( object ) == (void *)( 0 ) ) {
    strcpy( ret, "<null pointer exception>" );
    return W_DONE;
  }
  cl = OBJECT_CLASS( object );

  if( CLASS_IS_ARRAY( cl ) ) {
    slots addr, slength, sindex;
    int length, index, i, c;
    addr.v.taddr = object;
    load_offset_int( &slength, &addr, object_array_length );
    length = slength.v.tint;

    if( ( strcmp( head, "length" ) == 0 ) && !pos ) {
      sprintf( ret, "%d", length );
      return W_DONE;
    }      

    /* now look at the index we've been given */
    for( i = 0; i < strlen( head ); i++ )
      if( !isdigit( head[ i ] ) && !isspace( head[ i ] ) ) {
	strcpy( ret, "<bad index>" );
	return W_DONE;
      }
    index = atoi( head );
    if( ( index < 0 ) || ( index >= length ) ) {
      strcpy( ret, "<index out of range>" );
      return W_DONE;
    }
    sindex.v.tint = index;

    c = CLASS_CNAME( cl )[ 1 ];
    if( ( c != 'L' ) && ( c != '[' ) && pos ) {
      return W_UNKNOWN;
    }

    switch( c ) {
    default: /* we have a reference, is it a string? */
      if( !pos ) {
        /* test for String */
	Utf8Const *stringconst;
	stringconst = utf8ConstNew( "[Ljava/lang/String;", -1 );
        if( !stringconst )
	  return W_MEMORY;
        if( utf8ConstEqual( cl->name, stringconst ) ) {
          jstring jstr;
          char str[ 1027 ];
          load_offset_scaled_ref( &s, &addr, &sindex, object_array_offset );
          jstr = (jstring)( s.v.taddr );
	  stringJava2CBuf( jstr, str, 1026 );
          if( strlen( str ) + 2 < 1024 ) {
            strcpy( ret, "\"" );
            strcat( ret, str );
            strcat( ret, "\"" );
          } else
            strcpy( ret, "<string too long>" );
          utf8ConstRelease( stringconst );
          return W_DONE;
        } else {
          utf8ConstRelease( stringconst );
	  load_offset_scaled_ref( &s, &addr, &sindex, object_array_offset );
	  if( (void *)( s.v.taddr ) == 0 )
	    return W_NULL;
	  else {
	    getToString( ret, s.v.taddr,
			 callback );
	    return W_DONE;
	  }
        }
      } else if( pos ) {
	load_offset_scaled_ref( &s, &addr, &sindex, object_array_offset );
        return lookup_field( (Hjava_lang_Object *)( s.v.taddr ), tail, ret, 
			     pc, meth, lcl, callback );
      }
    case 'B':
      load_offset_scaled_byte( &s, &addr, &sindex, object_array_offset );
      sprintf( ret, "%d", s.v.tint );
      return W_DONE;
    case 'C':
      load_offset_scaled_char( &s, &addr, &sindex, object_array_offset );
      ch = s.v.tint;
      if( ( ch < 0x100 ) && isprint( (char)ch ) )
	sprintf( ret, "'%c'", (char)ch );
      else
	sprintf( ret, "0x%x", ch );
      return W_DONE;
    case 'D':
      load_offset_scaled_double( &s, &addr, &sindex, object_array_offset );
      sprintf( ret, "%g", s.v.tdouble );
      add0( ret );
      return W_DONE;
    case 'F':
      load_offset_scaled_float( &s, &addr, &sindex, object_array_offset ); 
      sprintf( ret, "%g", s.v.tfloat );
      add0( ret );
      return W_DONE;
    case 'I':
      load_offset_scaled_int( &s, &addr, &sindex, object_array_offset );
      sprintf( ret, "%d", s.v.tint );
      return W_DONE;
    case 'J':
      load_offset_scaled_long( &s, &addr, &sindex, object_array_offset );
      sprintf( ret, "%lld", s.v.tlong );
      return W_DONE;
    case 'S':
      load_offset_scaled_short( &s, &addr, &sindex, object_array_offset );
      sprintf( ret, "%d", s.v.tint );
      return W_DONE;
    case 'Z':
      load_offset_scaled_byte( &s, &addr, &sindex, object_array_offset );
      if( s.v.tint == 0 )
	strcpy( ret, "false" );
      else
	strcpy( ret, "true" );
      return W_DONE;
    }
  }

  /* now create a utf8 for the head */
  utf8 = utf8ConstNew( head, -1 );
  if( !utf8 )
    return W_MEMORY;

  nf = cl->nfields;
  nsf = cl->nsfields;
  for( j = 0; j < nf; j++ ) {
    if( utf8ConstEqual( ( &( cl->fields )[ j ] )->name, utf8 ) ) {
      fld = &( cl->fields )[ j ];
      if( FIELD_ISREF( fld ) && !pos ) { 
        /* test for String */
        errorInfo info;
	Hjava_lang_Class *stringclass = lookupClass( "java/lang/String", NULL, &info );
	if( fld->type == stringclass ) {
	  jstring jstr;
	  char str[ 1027 ];
	  if( FIELD_IS_STATIC( j ) ) 
            load_addr_ref( &s, FIELD_ADDRESS( fld ) );
	  else
	    load_o_ref( &s, object, FIELD_BOFFSET( fld ) );
	  jstr = (jstring)( s.v.taddr );
	  stringJava2CBuf( jstr, str, 1026 );
	  if( strlen( str ) + 2 < 1024 ) {
	    strcpy( ret, "\"" );
	    strcat( ret, str );
	    strcat( ret, "\"" );
	  } else
	    strcpy( ret, "<string too long>" );
	  utf8ConstRelease( utf8 );
	  return W_DONE;
	} else {
	  utf8ConstRelease( utf8 );
	  if( FIELD_IS_STATIC( j ) ) 
            load_addr_ref( &s, FIELD_ADDRESS( fld ) );
	  else
	    load_o_ref( &s, object, FIELD_BOFFSET( fld ) );
	  if( (void *)( s.v.taddr ) == 0 )
	    return W_NULL;
	  else {
	    getToString( ret, s.v.taddr,
			 callback );
	    return W_DONE;
	  }
	}
      } else if( FIELD_ISPRIM( fld ) && !pos ) {
	if( FIELD_IS_STATIC( j ) ) {
	  int c;
	  switch( CLASS_PRIM_SIG( fld->type ) ) {
	  case 'B': 
	    load_addr_byte( &s, FIELD_ADDRESS( fld ) );
	    sprintf( ret, "%d", s.v.tint );
	    utf8ConstRelease( utf8 );
            return W_DONE; 
	  case 'C':
	    load_addr_char( &s, FIELD_ADDRESS( fld ) );
            c = s.v.tint;
            if( ( c < 0x100 ) && isprint( (char)c ) )
              sprintf( ret, "'%c'", (char)c );
            else
              sprintf( ret, "0x%x", c );
            utf8ConstRelease( utf8 );
            return W_DONE;	
	  case 'D':
	    load_addr_double( &s, FIELD_ADDRESS( fld ) );
	    sprintf( ret, "%g", s.v.tdouble );
	    add0( ret );
	    utf8ConstRelease( utf8 );
	    return W_DONE;
	  case 'F':
	    load_addr_float( &s, FIELD_ADDRESS( fld ) );
	    sprintf( ret, "%g", s.v.tfloat );
	    add0( ret );
	    utf8ConstRelease( utf8 );
	    return W_DONE;
	  case 'I':
	    load_addr_int( &s, FIELD_ADDRESS( fld ) );
	    sprintf( ret, "%d", s.v.tint );
	    utf8ConstRelease( utf8 );
	    return W_DONE;
	  case 'J':
	    load_addr_long( &s, FIELD_ADDRESS( fld ) );
	    sprintf( ret, "%lld", s.v.tlong );
	    utf8ConstRelease( utf8 );
	    return W_DONE;
	  case 'S':
	    load_addr_short( &s, FIELD_ADDRESS( fld ) );
	    sprintf( ret, "%d", s.v.tint );
	    utf8ConstRelease( utf8 );
	    return W_DONE;
	  case 'Z':
	    load_addr_byte( &s, FIELD_ADDRESS( fld ) );
	    if( s.v.tint == 0 )
              strcpy( ret, "false" );
            else
              strcpy( ret, "true" );
            utf8ConstRelease( utf8 );
            return W_DONE;	        
	  }
	} else {
	  int c;
	  switch( CLASS_PRIM_SIG( fld->type ) ) {
	  case 'B':
	    load_o_byte( &s, object, FIELD_BOFFSET( fld ) );
	    sprintf( ret, "%d", s.v.tint );
	    utf8ConstRelease( utf8 );
	    return W_DONE;
          case 'C':
            load_o_char( &s, object, FIELD_BOFFSET( fld ) );
            c = s.v.tint;
            if( ( c < 0x100 ) && isprint( (char)c ) )
              sprintf( ret, "'%c'", (char)c );
            else
              sprintf( ret, "0x%x", c );
            utf8ConstRelease( utf8 );
            return W_DONE;
          case 'D':
            load_o_double( &s, object, FIELD_BOFFSET( fld ) );
            sprintf( ret, "%g", s.v.tdouble );
	    add0( ret );
            utf8ConstRelease( utf8 );
            return W_DONE;
          case 'F':
            load_o_float( &s, object, FIELD_BOFFSET( fld ) );
            sprintf( ret, "%g", s.v.tfloat );
	    add0( ret );
            utf8ConstRelease( utf8 );
            return W_DONE;
          case 'I':
            load_o_int( &s, object, FIELD_BOFFSET( fld ) );
            sprintf( ret, "%d", s.v.tint );
            utf8ConstRelease( utf8 );
            return W_DONE;
          case 'J':
            load_o_long( &s, object, FIELD_BOFFSET( fld ) );
            sprintf( ret, "%lld", s.v.tlong );
            utf8ConstRelease( utf8 );
            return W_DONE;
          case 'S':
            load_o_short( &s, object, FIELD_BOFFSET( fld ) );
            sprintf( ret, "%d", s.v.tint );
            utf8ConstRelease( utf8 );
            return W_DONE;
          case 'Z':
            load_o_byte( &s, object, FIELD_BOFFSET( fld ) );
            if( s.v.tint == 0 )
              strcpy( ret, "false" );
            else
              strcpy( ret, "true" );
            utf8ConstRelease( utf8 );
            return W_DONE;
	  }
	}
      } else if( FIELD_ISREF( fld ) ) {
	if( FIELD_IS_STATIC( j ) )
	  load_addr_ref( &s, FIELD_ADDRESS( fld ) );
	else
	  load_o_ref( &s, object, FIELD_BOFFSET( fld ) );
	utf8ConstRelease( utf8 );
        return lookup_field( (Hjava_lang_Object *)( s.v.taddr ), tail, ret, 
			     pc, meth, lcl, callback );
      }
    }
  }

  utf8ConstRelease( utf8 );
  return W_UNKNOWN;
}

/* rather silly little function */
int cont( int i ) {
  return( ( i == W_UNKNOWN ) || ( i == W_SCOPE ) || ( i == W_LVT ) );
}

/* Lookup a variable, return is one of W_DONE, W_LVT, ... */
int lookup_variable( uintp pc, methods *meth, slots *lcl,   
                     char path[ 256 ], char *ret, callbackThread* callback ) {
  int retn = W_UNKNOWN;
  Utf8Const *utf8;
  int r;

  /* Separate head and tail */
  char *pos;
  char head[ 256 ] = "";
  char tail[ 256 ] = "";
  pos = shft( path, head, tail,
	      pc, meth, lcl, callback );

  /* now create a utf8 for the head */
  utf8 = utf8ConstNew( head, -1 );
  if( !utf8 )
    return W_MEMORY;

  /* First search the local variable table if loaded */
  if( meth->localvars == 0 )
    retn = W_LVT;
  else {
    int j;
    int match = -1, latest = 0;

    /*
      PSEUDOCODE:
      -----------
      match = -1;
      latest = 0;
      for( each entry ) {
        if( correct name ) {
          if( in scope && startpc >= latest ) {
            match = entry number
            latest = startpc
          }
	} else if( !in scope )
          retn = W_SCOPE;
      }
      if( match != -1 ) {
        if( no dot )
          switch on type and return
        else if( unmatched type )
          return lookup_field
      }
    */

    for( j = 0; j < meth->localvars->length; j++ ) {
      if( utf8ConstEqual( GET_CONST( meth, meth->localvars->entry[ j ].name_index ),
			  utf8 ) ) {
	int start = meth->localvars->entry[ j ].start_pc;
	int end = start + meth->localvars->entry[ j ].length;

	if( ( pc >= start ) && ( pc <= end ) && ( start >= latest ) ) {
	  match = j;
	  latest = start;
	}
	else if( ( pc < start ) || ( pc > end ) )
	  retn = W_SCOPE;
      }
    }
    
    if( match != -1 ) {
      char d = get_descriptor_type( GET_CONST( meth, meth->localvars->entry[ match ].descriptor_index ) );
      if( !pos ) {
	switch( d ) {
	case 'X': 
	  utf8ConstRelease( utf8 );
	  return W_MEMORY;
	case 'B': case 'I': case 'S':
	  sprintf( ret, "%d", ( local( meth->localvars->entry[ match ].index ) )[0].v.tint );
	  utf8ConstRelease( utf8 );
	  return W_DONE;
	case 'Z':
	  if( ( local( meth->localvars->entry[ match ].index ) )[0].v.tint == 0 )
	    strcpy( ret, "false" );
	  else
	    strcpy( ret, "true" );
	  utf8ConstRelease( utf8 );
	  return W_DONE;
	case 'J':
	  sprintf( ret, "%lld", 
		   ( local( meth->localvars->entry[ match ].index ) )[0].v.tlong );
	  utf8ConstRelease( utf8 );
	  return W_DONE;
	case 'F':
	  sprintf( ret, "%g", 
		   ( local( meth->localvars->entry[ match ].index ) )[0].v.tfloat );
	  add0( ret );
	  utf8ConstRelease( utf8 );
	  return W_DONE;
	case 'D':
	  sprintf( ret, "%g", 
                   ( local( meth->localvars->entry[ match ].index ) )[0].v.tdouble );
	  add0( ret );
	  utf8ConstRelease( utf8 );
	  return W_DONE;
	case 'C': {
	  int c = ( local( meth->localvars->entry[ match ].index ) )[0].v.tint;
	  if( ( c < 0x100 ) && isprint( (char)c ) ) 
	    sprintf( ret, "'%c'", (char)c );
	  else
	    sprintf( ret, "0x%x", c );
	  utf8ConstRelease( utf8 );
	  return W_DONE;
	}
	case 'T': {
	  jstring jstr = (jstring)( ( local( meth->localvars->entry[ match ].index ) )[0].v.taddr );
	  char str[ 1027 ];
	  stringJava2CBuf( jstr, str, 1026 );
	  if( strlen( str ) + 2 < 1024 ) {
	    strcpy( ret, "\"" );
	    strcat( ret, str );
	    strcat( ret, "\"" );
	  } else 
	    strcpy( ret, "<string too long>" );
	  utf8ConstRelease( utf8 );
	  return W_DONE;
	}
	default:
	  utf8ConstRelease( utf8 );
	  if( (void *)( ( local( meth->localvars->entry[ match ].index ) )[0].v.taddr ) == 0 )
	    return W_NULL;
	  else {
	    getToString( ret, ( ( local( meth->localvars->entry[ match ].index ) )[0].v.taddr ),
			 callback );
	    return W_DONE;
	  }
	}
      } else if ( d == 'U' ) { 
	utf8ConstRelease( utf8 );
	return lookup_field( (Hjava_lang_Object *)( ( local( meth->localvars->entry[ match ].index ) )[0].v.taddr ), 
			     tail, ret, pc, meth, lcl, callback );
      } 
    }
  }

  /* now see if this.<path> is a var, and failing this whether it is a var of this
     class and failing this whether head is a class and failing this, whether
     the first few . separated components (with .'s changed to /) are a class */
  if( strcmp( "this", head ) != 0 ) {
    char newpath[ 256 ];
    strcpy( newpath, "this." );
    strcat( newpath, path );
    r = lookup_variable( pc, meth, lcl, newpath, ret, callback );
    if( !cont( r ) ) {
      utf8ConstRelease( utf8 );
      return r;
    } else {
      r = lookup_class_field( meth->class, path, ret, 
			      pc, meth, lcl, callback );   
      if( !cont( r ) ) {
	utf8ConstRelease( utf8 );
	return r;
      } else {
        /* now find out if head=~=utf8 is a class and if so pass it onto 
         * lookup_class_variable( class, tail, ret ) */
	errorInfo info;
	Hjava_lang_Class* cl;
	for( ; ; ) {
	  cl = lookupClass( head, NULL, &info );
	  if( cl != 0 ) {
	    utf8ConstRelease( utf8 ); 
	    r = lookup_class_field( cl, tail, ret, 
				    pc, meth, lcl, callback );
	    if( !cont( r ) ) {
	      utf8ConstRelease( utf8 );
	      return r;
	    } else {
	      char tailhead[ 256 ];
	      char tailtail[ 256 ];
	      char *p = shft( tail, tailhead, tailtail,
			      pc, meth, lcl, callback );
	      if( p ) {
		strcat( head, "/" );
		strcat( head, tailhead );
		strcpy( tail, tailtail );
		continue;
	      } else break;
	    }
	  } else {
            char tailhead[ 256 ];
            char tailtail[ 256 ];
            char *p = shft( tail, tailhead, tailtail,
			    pc, meth, lcl, callback );
            if( p ) {
              strcat( head, "/" );
              strcat( head, tailhead );
              strcpy( tail, tailtail );
              continue;
	    } else break;
	  }
	}
      }
    }
  }  
  utf8ConstRelease( utf8 );  
  return retn;
}

/* END JJC (watches) */

/* START JJC (reverse simulation) */

/* the reverse simulation: more or less simulates what will happen during a step back,
 * but doesn't alter anything that's difficult to restore after the simulation and
 * before (if it is permitted) the real step back */

#define CHECK_VALID_PCBUFF { if( !(debug->cbufferpc_v) ) return -1; }  
#define CHECK_VALID_CBUFF  { if( !(debug->cbuffer_v) ) return -1; }
#define CHECK_VALID_CBUFF_NULL { if( !(debug->cbuffer_v ) ) return (methods*)0; }

#define SIM_REV (over_mode == R_OVERFIRST ? R_OVER : over_mode ) 

#define JJC_SIMULATION

#if defined(JJC_SIMULATION)

/* returns 1 (step back is ok), -1 (not ok), 0 (not sure, keep going).
 * If continuation is true, a step out of the beginning of a function
 * is simulated by recursively calling reverseVirtualmachine_simulation
 * for the parent method (in the real reversal we would just return).
 * If it is false we just return.
 */
int reverseVirtualMachine_simulation( methods* meth, debugThread* debug, callbackThread* callback, 
				      int over_mode, bool continuation, bool ignore_first,
				      volatile vmException* mjbuf ) {

  /* the mjbuf is not used anymore */

  bytecode *code = (bytecode*)meth->c.bcode.code;
  bool normalInstruction;
  int32 lastpc; 
  int32 l; /* line we are reversing over */
  int32 pc;
  int ret;

  slots tmp[1];
  slots mtable[1];
  
  int32 idx;
  callInfo cinfo;
  errorInfo einfo;

  D_FLS( fprintf( stderr, "............\nreverseVirtualMachine_simulation: meth: %s, over_mode = %d, cont = %d, ignore_first = %d\n",
		  meth->name->data, over_mode, (int)continuation, (int)ignore_first ) );

  CHECK_VALID_PCBUFF; pc = rcbuffpc( debug );

  lastpc = 0; /* this initial value is never looked at */

  assert( !( ignore_first && ( pc & THROW_MARKER ) ) );

  l = currentPoint( pc, meth );

  if( ignore_first ) {
    lastpc = pc;
    D_FLS( fprintf( stderr, "Before ignore_first: cbuffp = %d\n", debug->cbuffer_p - debug->cbuffer_s ) );
    CHECK_VALID_PCBUFF; pc = rcbuffpc( debug ); 
    if( !rcbuffsim( debug ) ) {
      D_FLS( fprintf( stderr, "......Sim RETURN: buffer ran out on skip first\n" ) );
      return -1;  
    }
    D_FLS( fprintf( stderr, "After ignore_first: cbuffp = %d\n", debug->cbuffer_p - debug->cbuffer_s ) );
  }

  while( ( debug->breakpoint )  ||
	 ( l == currentPoint( pc, meth ) ) || 
	 ( over_mode == R_OVER ) ||
	 ( pc & FILE_BIT ) ||
	 ( meth->deepLog == true ) ) {
   
    normalInstruction = true;

    if( pc == START_OF_CATCH ) {
      /* read off the START_OF_CATCH and discard method pointer */
      CHECK_VALID_PCBUFF; rcbuffpc( debug ); 
      CHECK_VALID_PCBUFF; pc = rcbuffpc( debug );
    }

    /* do reversal for insn at pc */
    D_FLS( fprintf( stderr, "Simulation acting on pc = %d (0x%x), cbuffp = %d, line no = %d\n", pc, pc, debug->cbuffer_p - debug->cbuffer_s, l ) );

    if( pc & THROW_MARKER ) {
      int32 type = pc;
      methods* new_meth;
      int32 new_pc;
      normalInstruction = false;

      CHECK_VALID_PCBUFF; new_meth = (methods*)rcbuffpc( debug ); 
      CHECK_VALID_PCBUFF; new_pc = rcbuffpc( debug ); /* discard the program counter (never any use here) */
      
      if( !rcbuffsim( debug ) ) {
	D_FLS( fprintf( stderr, "......\nFailed at THROW MARKER check\n" ) );
	return -1;
      }

      D_FLS( fprintf( stderr, "Simulating exception: method %s, type %d, pcbuff = %d\n", 
		      meth->name->data, (int)(type==THROW_FIRST), debug->cbuffer_p - debug->cbuffer_s  ) );


      if( !( type & THROW_LAST_MASK ) ) {
	ret = reverseVirtualMachine_simulation( new_meth, debug, callback, 
						( type & THROW_LAST_MASK ? over_mode : SIM_REV ), 
						false, type & THROW_FIRST_MASK, 0 );
	if( ret != 0 ) {
	  D_FLS( fprintf( stderr,  "......\nret from Throw Marker call = %d\n", ret ) );
	  return ret;
	}
      }
    }
    else if( pc == START_OF_METHOD ) {
      methods* new_meth;
      int32 new_meth_pc;
      CHECK_VALID_PCBUFF;
      new_meth_pc = rcbuffpc( debug );
      new_meth = (methods*)( new_meth_pc );
      assert( new_meth );
      if( continuation ) {
	D_FLS( fprintf( stderr, "Simulation Continuation...\n" ) );
	return reverseVirtualMachine_simulation( new_meth, /* mjbuf->prev->meth, */
						 debug, callback,
						 R_INTO, true, false, mjbuf->prev );
      }
      else {
	D_FLS( fprintf( stderr, "......Simulation Return 0.\n" ) );
	return 0;
      }
    }
    else if( pc == END_OF_METHOD ) { 
      /* should never happen */
      fprintf( stderr, "ABORT: unexpected EOM in simulation." );
      assert( 0 );
    } 
    else if( pc & METHOD_OFFSET ) {
      int suc;
      int32 end_meth;
      normalInstruction = false;
      pc = pc & 0xFFFF;
      CHECK_VALID_PCBUFF;
      end_meth = rcbuffpc( debug );
      if( end_meth != END_OF_METHOD ) {
	if( code[ pc ] == INVOKEVIRTUAL ||
	    code[ pc ] == INVOKEINTERFACE ||
	    code[ pc ] == INVOKESPECIAL ) 
	  if( !rcbuffs_test( debug ) ) {
	    D_FLS( fprintf( stderr, "......Simulation Return -1 (after non-valid)\n" ) );
	    return -1;  
	  }
	normalInstruction = true;
      } else {

	switch( code[ pc ] ) {
	default: {
	  fprintf( stderr, "Unexpected instruction for rev fn\n" );
	  assert( 0 );
	  break;
	}
#include "invokesim.def"
	}

	if( suc != 0 ) {
	  D_FLS( fprintf( stderr, "Simulation Return: Decided after invokesim = %d\n", suc ) );
	  return suc;
	}
	 
      }
    }
    if( normalInstruction ) {
      if( !rcbuffsim( debug ) ) {
	D_FLS( fprintf( stderr, "Simulation Return: Failed on normal instruction\n" ) );
	return -1;
      }
    }

    lastpc = pc;
    CHECK_VALID_PCBUFF; pc = rcbuffpc( debug );

    if( l == ( debug->bytecode || !meth->haveSource ? RECTIFY_FAIL : 0 ) )
      l = currentPoint( pc, meth );

    if( currentPoint( pc, meth ) != ( debug->bytecode || !meth->haveSource ? RECTIFY_FAIL : 0 ) )
      if( debug->breakpoint &&
	  isBreakpoint( currentPoint( pc, meth ), callback, meth, currentIsPc( pc, meth ) ) ) {
	debug->breakpoint = 0;
	l = currentPoint( pc, meth );
	over_mode = R_OVERFIRST;
      }
  }
  
  wcbuffpc( pc, debug );
  D_FLS( fprintf( stderr, "......Simulation Return: reached destination\n" ) );
  return 1;  
}

/* Return true if the step back (into/over/bpt) is possible, false if it is not.
 * Wrapper for the simulation: saves and restores values changed by the simulation.
 * The breakpoint flag is dealt with by the forwards code.
 */
bool doSimulation( methods* meth, debugThread* debug, callbackThread* callback, 
		   int over_mode, volatile vmException* mjbuf ) {
  slots* cbuffer_p_save = debug->cbuffer_p;
  slots* cbuffer_v_save = debug->cbuffer_v;
  uint16* cbuffercount_p_save = debug->cbuffercount_p;
  int8* cbufferobj_p_save = debug->cbufferobj_p;
  uintp* cbufferpc_p_save = debug->cbufferpc_p;
  uintp* cbufferpc_v_save = debug->cbufferpc_v;

  int result;
  
  result = reverseVirtualMachine_simulation( meth, debug, callback, 
					     over_mode, true, false, mjbuf );

  debug->bank = -1;

  debug->cbuffer_p_test = debug->cbuffer_p;
  debug->cbuffer_v_test = debug->cbuffer_v;
  debug->cbufferpc_p_test = debug->cbufferpc_p;
  debug->cbufferpc_v_test = debug->cbufferpc_v;
  debug->test_on_next_stop = 1;

  debug->cbuffer_p = cbuffer_p_save;
  debug->cbuffer_v = cbuffer_v_save;
  debug->cbuffercount_p = cbuffercount_p_save;
  debug->cbufferobj_p = cbufferobj_p_save;
  debug->cbufferpc_p = cbufferpc_p_save;
  debug->cbufferpc_v = cbufferpc_v_save;

  return result == 1; 
}

#else

#warning "No Backwards Simulation!"
/* the following provides a way of 'switching off' simulation testing */

bool doSimulation( methods* meth, debugThread* debug, callbackThread* callback, 
		   int over_mode, volatile vmException* mjbuf ) {
  return true;
}

#endif

/* END JJC (reverse simulation) */

/* START JJC (real reversal) */

/* return F_BEG, F_END or F_CATCH. */
/* ignore_first is a hack to be used when reversing exceptions:
 * the first instruction is the one that faulted and should not
 * be reversed.   We believe that all instructions only fault
 * _before_ doing things that might upset this scheme
 */

int reverseVirtualMachine( methods* meth, slots* lcl, slots** spp, uintp* npcp, 
			   slots* retval, volatile vmException *mjbuf, Hjava_lang_Thread *tid,
			   debugThread* debug, callbackThread* callback, 
			   char* sourceFile, int8** opsobjsp, int8* lvtobjs,
			   int over_mode, int32* cl, bool* seen_new_line, bytecode* code,
			   int8 thr, bool ignore_first ) {
  /* Misc machine variables */
  slots tmp[1];
  slots mtable[1];
  
  /* Variables used directly in the machine */
  int32 idx;
  jint low;
  
  /* Call, field and creation information */
  callInfo cinfo;
  fieldInfo finfo;
  errorInfo einfo;

  int i, n;

  slots *sp = *spp;
  uintp npc = *npcp;
  int8* opsobjs = *opsobjsp;
  bool normalInstruction;
  int32 lastpc;
  int32 l; /* line we are reversing over */
  int32 pc = rcbuffpc( debug );
  lastpc = 0; /* this initial value is never looked at */

  D_FL( fprintf( stderr, "~~~~~~~~~~~~\nReverse Virtual Machine: %s: ignorefirst %d, whole_fn = %d\n", meth->name->data, ignore_first, over_mode ) );

  l = currentPoint( pc, meth );

  /* if this is the reversal of an exception then skip the instruction that faulted,
   * the counter scheme for reverse simulation has here found a totally unexpected use. */
  if( ignore_first ) {
    lastpc = pc;
    pc = rcbuffpc( debug ); 
    wipeOffLastInstruction( debug );
  }

  while( ( debug->breakpoint )  ||
	 ( l == currentPoint( pc, meth ) ) ||
	 ( over_mode == R_OVER ) ||
	 ( pc & FILE_BIT ) ||
	 ( meth->deepLog == true ) ) {
    normalInstruction = true;

    D_FL( fprintf( stderr, "Reverse Virtual Machine: acting on pc = %d (0x%x), cbuffp = %d, line no = %d\n", pc, pc, debug->cbuffer_p - debug->cbuffer_s, l ) );

    /* do reversal for insn at pc */
    if( pc & FILE_BIT ) {
      switch( pc ) {
      case FILE_OPEN:  fileOpenReverse( debug ); break;
      case FILE_CLOSE: fileCloseReverse( debug ); break;
      case FILE_READ:  fileReadReverse( debug ); break;
      case FILE_WRITE: fileWriteReverse( debug ); break;
      case FILE_LSEEK: fileLseekReverse( debug ); break;
      default:
	fprintf( stderr, "Bad file flag on pc buffer\n" );
	assert( 0 );
      }
      wcbuff_decrement( debug );
      normalInstruction = false;
      pc = rcbuffpc( debug );

      if( l == ( debug->bytecode || !meth->haveSource ? RECTIFY_FAIL : 0 ) )
	l = currentPoint( pc, meth );

      continue;
    }
    else if( pc == START_OF_METHOD || pc == START_OF_CATCH ) {
      rcbuffpc( debug ); /* discard the method pointer (used for the simulation only) */
      *spp = sp;
      *npcp = npc;
      *opsobjsp = opsobjs;
      D_FL( fprintf( stderr, "~~~~~~Reverse Virtual Machine: meth %s, return %x \n", meth->name->data, pc ) );
      return ( pc == START_OF_METHOD ? F_BEG : F_CATCH );
    }
    else if( pc == END_OF_METHOD ) { 
      /* should never happen */
      fprintf( stderr, "ABORT: unexpected EOM" );
      assert( 0 );
    } 
    else if( pc & METHOD_OFFSET ) {
      int8 beg = F_BEG;
      uintp last;

      normalInstruction = false;
      pc = pc & 0xFFFF;
      last = rcbuffpc( debug );
      if( last != END_OF_METHOD ) {
	if( code[ pc ] == INVOKEVIRTUAL ||
	    code[ pc ] == INVOKEINTERFACE ||
	    code[ pc ] == INVOKESPECIAL ) 
	  rcbuffs( debug );
	if( last & FILE_BIT ) {
	  lastpc = pc;
	  pc = last;	  
	  continue;
	} else
	  normalInstruction = true;
      } else {
	switch( code[ pc ] ) {
	default: {
	  fprintf( stderr, "Unexpected instruction for rev fn\n" );
	  assert( 0 );
	  break;
	}
#include "invoke.def"
	}
	if( beg == F_END ) {
	  wcbuffpc( pc + METHOD_OFFSET, debug );
	  npc = pc + insnLen[code[pc]];
	  *cl = ( debug->bytecode ? currentPc( rectify( pc ) ) : ( meth->haveSource ? currentLine( pc, meth ) : RECTIFY_FAIL ) );
	  *seen_new_line = false;
	  *spp = sp;
	  *npcp = npc;
	  *opsobjsp = opsobjs;
	  D_FL( fprintf( stderr, "~~~~~~Reverse Virtual Machine: F_END from invoke.def; method %s\n", meth->name->data ) );
	  return F_END;
	} else
	  l = currentPoint( pc, meth ); /* Experiment! */
      }
    } 
    if( normalInstruction ) {
      assert(pc < meth->c.bcode.codelen); /* just in case */
      npc = pc + insnLen[code[pc]];       /* needed for WIDE */
      
      switch (code[pc]) {
      default:
	restorestack( code[ pc ] );
	break;
#include "reverse.def"
      }
    }
    if( normalInstruction ) {
      wcbuff_decrement( debug );
      mjbuf->pc = pc;
      mjbuf->opsobjs = opsobjs;
      mjbuf->opstop = sp;
    }

    lastpc = pc;
    pc = rcbuffpc( debug );

    if( l == ( debug->bytecode || !meth->haveSource ? RECTIFY_FAIL : 0 ) )
      l = currentPoint( pc, meth );

    /*if( meth->haveSource ) {*/
    if( currentPoint( pc, meth ) != ( debug->bytecode || !meth->haveSource ? RECTIFY_FAIL : 0 ) )
      if( debug->breakpoint && isBreakpoint( currentPoint( pc, meth ), callback, meth, currentIsPc( pc, meth ) ) ) { 
        debug->breakpoint = 0;
        l = currentPoint( pc, meth );
        over_mode = R_OVERFIRST;
      }
      /*}*/
  }

  wcbuffpc( pc, debug );
  npc = pc = lastpc;
  *cl = RECTIFY_FAIL;
  *seen_new_line = false;
  *spp = sp;
  *npcp = npc;
  *opsobjsp = opsobjs;
  D_FL( fprintf( stderr, "~~~~~~Reverse Virtual Machine: Return: reached destination; method %s\n", meth->name->data ) );
  return F_END;  
}

/* END JJC (real reversal) */

/* START JJC */

void generateNames( methods* meth ) {
  char* filePtr;
  char* classFilePtr;
  char sourceFileWithSpace[ 512 ] = " ";
  char* slashloc;
  char classnameWithSpace[ 512 ] = " ";

  /* Investigate the filename */
	      
  meth->fullSourceFile = (char*)gc_malloc( ( meth->class->name ? strlen( meth->class->name->data ) : 0 )
					   + strlen( (char*)CLASS_SOURCEFILE( meth->class ) ) + 2, 
					   GC_ALLOC_FIXED );
  if( !meth->fullSourceFile )
    fprintf( stderr, "No room for source file name.\n" );
	      
  if( meth->class->name )
    strcpy( meth->fullSourceFile, meth->class->name->data );
  else
    strcpy( meth->fullSourceFile, "" );
  for( slashloc = meth->fullSourceFile; *slashloc != 0; slashloc++ )
    if( *slashloc == '.' )
      *slashloc = '/';
	      
  slashloc = strrchr( meth->fullSourceFile, '/' );
  if( slashloc )
    *( slashloc + 1 ) = 0;
  else
    *( meth->fullSourceFile ) = 0;
  strcat( meth->fullSourceFile, (char*)CLASS_SOURCEFILE( meth->class ) );
	      
  strcat( sourceFileWithSpace, meth->fullSourceFile );
	      
  /* Investigate the classname */
	      
  strcat( classnameWithSpace, meth->class->name->data );
  for( slashloc = classnameWithSpace; *slashloc != 0; slashloc++ )
    if( *slashloc == '$' )
      *slashloc = '.';
  strcat( classnameWithSpace, ".class" );

  /* construct the method's bytecode breakpoint signature
   * of the form: <classname>.class:<methodname><methodsig>
   */

  if( meth->name && meth->parsed_sig )
    if( meth->parsed_sig->signature ) {
      meth->bytecodeBreakpointSig = (char*)gc_malloc(
						     strlen( classnameWithSpace ) + strlen( meth->name->data )
						     + strlen( meth->parsed_sig->signature->data ) + 3,
						     GC_ALLOC_FIXED );
      strcpy( meth->bytecodeBreakpointSig, classnameWithSpace );
      strcat( meth->bytecodeBreakpointSig, ":" );
      strcat( meth->bytecodeBreakpointSig, meth->name->data );
      strcat( meth->bytecodeBreakpointSig,
	      meth->parsed_sig->signature->data );
      ( meth->bytecodeBreakpointSig )++; /* deal with the space at the start */
    }

  meth->haveSource = ( ( filePtr = strstr( debug_file, sourceFileWithSpace ) ) != 0 )
    && meth->lines; /* = not == intentional. */
  meth->haveBytecode = ( ( classFilePtr = strstr( debug_file, classnameWithSpace ) ) != 0 ); /* ditto */
  if( meth->haveSource ) 
    meth->fileOffset = filePtr - debug_file;
  if( meth->haveBytecode )
    meth->bytecodeFileOffset = classFilePtr - debug_file;  
}

/* END JJC */

int8 runVirtualMachine(methods *meth, slots *lcl, slots *sp, uintp npc, slots *retval, volatile vmException *mjbuf, Hjava_lang_Thread *tid, int8 reverse_method, int8 thr ) { /* JJC (extra parameters) */
  bytecode *code = (bytecode*)meth->c.bcode.code;

  /* Misc machine variables */
  jlong lcc;
  jlong tmpl;
  slots tmp[1];
  slots tmp2[1];
  slots mtable[1];

  /* Variables used directly in the machine */
  int32 idx;
  jint low;
  jint high;

  /* Call, field and creation information */
  callInfo cinfo;
  fieldInfo finfo;
  Hjava_lang_Class* crinfo;
  errorInfo einfo;

  /* START JJC (variables and deal with method entry) */
  int8* opsobjs = 0; 
  int8* lvtobjs = 0;
  slots *osp = sp;  /* so we can tell how deep the stack is */ 
  bool valid_log = false; /* do we want to log? */
  int32 cl = 0; 
  int32 cline = 0;
  bool seen_new_line = true;
  bool is_main = false, should_bpt = false;
  callbackThread* callback = 0;
  debugThread* debug = 0;
  int oldopsobjs = 0; /* init this to stop the compiler getting worried */

  bool is_breakpoint;
  bool running_over = ( reverse_method == T_FWDS_OVER );
  int i;
  bool is_catch = ( reverse_method & T_CATCHOFFSET ) != 0; /* is this method a catch invocation? */

  int ret;

  /* STEP 1: Is this the method to trigger "debug" mode */

  if( debugging_enabled == 0 ) {
    debugging_enabled = ( (int)Kaffe_JavaVMArgs[ 0 ].debugJJC == 1 ) ? 1 : -1;
    pragmaJJC = ( (int)Kaffe_JavaVMArgs[ 0 ].pragmaJJC );
  }

  if( !cb_set )
    if( debugging_enabled == 1 ) 
      if( meth )
	if( strcmp( meth->class->name->data, CCBRIDGE ) == 0 )
	  if( strcmp( meth->name->data, "triggerInitialValueRequest" ) == 0 ) {
	    doSetup( );
	    /* no need to lock, only set once and no other methods running at that time */
	    cb_set = 1; 
	  }     	  
	
  if( cb_set ) {

    /* STEP 2: Is this method in a logged file? */
    if( meth->status == 0 ) {
      if( strcmp( "java/lang/Object", meth->class->name->data ) == 0 ) { 
	meth->status = -1;
      } else {
	generateNames( meth );          
	      
	if( !( meth->haveSource || meth->haveBytecode ) ) {
		
	  /* we exclude jjc/bdbj from deep logging as otherwise the user
	   * may not realise the implications of "deeplog *".
	   */
	  if( ( scanstar( meth->class->name->data, deep_list ) ) &&  
	      strncmp( "jjc/bdbj/", meth->class->name->data, 9 ) != 0 ) {
	    meth->status = 1;
	    meth->deepLog = true;
	  } else {
	    meth->status = -1;
	    meth->deepLog = false;
	  }   
	} else {
	  meth->deepLog = false;
	  meth->status = 1;
	  assert( meth->fullSourceFile );
        }          
	if( debugline ) {
	  if( meth->haveSource ) {
	    if( !lineInMethod( meth, debugline ) )
	      meth->is_init = false;
	    else if( strcmp( meth->fullSourceFile, initialfile ) != 0 )
	      meth->is_init = false;
	    else
	      meth->is_init = true;
	  } else
	    meth->is_init = false;
	}
      }
    }

    /* STEP 3: Have we just caught an exeception? */
	  
    if( is_catch ) 
      reverse_method &= T_MASK;
	  
    /* STEP 4: Make sure that we do not log the client: this means
     * that the debugger cannot debug any class in a package whose
     * name begins with "jjc/bdbj/"
     */
	  
    if( strncmp( "jjc/bdbj/", meth->class->name->data, 9 ) == 0 ) {
      if( strcmp( "invokeTarget", meth->name->data ) == 0 ) 
	reverse_method = T_SHALLOW;
      else
	reverse_method = T_DEEP;
    }
	  
    /* STEP 5: Is this the main method or initial method? */
	  
    if( ( meth->status > 0 ) && ( reverse_method == T_SHALLOW ) ) {
      if( debugline )
	is_main = meth->is_init;
      else {
	is_main = meth->deepLog != 1;
	should_bpt = is_main && strcmp( "main", meth->name->data ) != 0;
      }
      if( is_main ) 
	reverse_method = T_FWDS_INTO;
    }
	  
    /* STEP 6: Set valid_log to its initial value */
	  
    valid_log = ( meth->status > 0 ) && IS_LOGGED( reverse_method );
    mjbuf->is_main = is_main;
	  
    /* STEP 7: If we haven't yet created the pc lookup table do it */
	  
    if( valid_log && meth->status == 1 ) {
      if( meth->haveSource )
	createLookupTable( meth );
      meth->status = 2;
    }
	
    /* STEP 8: Get thread number or if we can't abandon logging */
	
    if( valid_log ) {
      if( is_main && thr == -1 )
	thr = enteredNewThread( tid );
      if( thr == -1 )
	valid_log = 0;
      mjbuf->miniThread = thr;
    }
	  	
    /* STEP 9: Set jump buffer values to default values */
	
    mjbuf->opstop = mjbuf->opsbase = sp;
    mjbuf->lclbase = lcl;
    mjbuf->logged = valid_log;
    mjbuf->opsobjs = 0;
    mjbuf->opsobjsbase = 0;
    mjbuf->lvtobjs = 0;
	
    if( valid_log ) {
	    
      /* STEP 10: If a catch adjust pointers as exception object already on stack */

      if( is_catch ) {
	--(mjbuf->opsbase);
	--osp;
      }
	    
      /* STEP 11: Initialize this method call for logging */
	    
      mjbuf->opsobjsbase = (int8*)alloca( meth->stacksz ) - ( is_catch ? 0 : 1 );
      mjbuf->lvtobjs = lvtobjs = (int8*)alloca( meth->localsz );
	    
      mjbuf->opsobjs = opsobjs = mjbuf->opsobjsbase;
      initLocalVariables( meth, mjbuf->lvtobjs );
	  
      if( is_catch )
	markisobj;
	    
      callback = callbackThreads[ thr ];
      debug = debugThreads[ thr ];
	  
      /* Now do the appropriate thing depending on the value of reverse_method */
	  
      D_FL( fprintf( stderr, "-----------\nRUN VIRTUAL MACHINE: entered method %s, reversemethod = %d\n", meth->name->data, reverse_method ) );
      if( IS_FORWARDS(reverse_method) && ( !is_main || is_catch ) ) {
	wcbuffpc( (uintp)(debug->last_method), debug );
	if( is_catch )
	  wcbuffpc( START_OF_CATCH, debug );
	else
	  wcbuffpc( START_OF_METHOD, debug );     
      }
	    
      if( IS_RETRY( reverse_method ) ) {
	uintp rec;
	restoreTopStack( mjbuf, debug );
	sp = mjbuf->opstop;
	opsobjs = mjbuf->opsobjs;

	rec = rcbuffpc( debug );
	rcbuffpc( debug ); /* discard method pointer */
	npc = rcbuffpc( debug );

	switch( rec ) {
	default:
	  fprintf( stderr, "Should be THROW_xxx, was %x\n", rec );
	  assert( 0 );
	case THROW_LAST:
	case THROW_MID:
	  if( !( rec & THROW_LAST_MASK ) )
	    if( reverse_method == T_RETRY_OVER_FIRST )
	      reverse_method = T_RETRY_OVER;
	  valid_log = 0;
	  break;
	case THROW_FIRST_LAST:
	case THROW_FIRST:
	  reverse_method = ( reverse_method == T_RETRY_OVER ? T_BACK_OVER : T_BACK_INTO );
	  ret = reverseVirtualMachine( meth, lcl, &sp, &npc, retval, mjbuf, tid,
				       debug, callback, meth->fullSourceFile, &opsobjs, lvtobjs,
				       reverse_method == T_BACK_OVER ? R_OVER : R_INTO,
				       &cl, &seen_new_line, code, thr, true /* TRUE! */ );
	  if( ret == F_BEG ) {
	    D_FL( fprintf( stderr, "------RUN VIRTUAL MACHINE: exit: ret == F_BEG\n" ) );
	    return F_BEG;
	  }
	  else if( ret == F_CATCH ) {
	    D_FL( fprintf( stderr, "------RUN VIRTUAL MACHINE: exit: ret == F_CATCH\n" ) );
	    fprintf( stderr, "I thought this couldn't happen (F_CATCH on THROW_FIRST).\n" );
	    return reverse_method == T_BACK_OVER ? F_RETRY_OVER : F_RETRY; /* a little unsure */
	  }
	  assert( reverse_method != T_BACK_OVER );
	  reverse_method = T_FWDS_INTO;
	  break;
	}
      } else if( reverse_method == T_BACK_INTO ) {
	ret = reverseVirtualMachine( meth, lcl, &sp, &npc, retval, mjbuf, tid,
				     debug, callback, meth->fullSourceFile, &opsobjs, lvtobjs, R_INTO, 
				     &cl, &seen_new_line, code, thr, false );
	if( ret == F_BEG ) {
	  D_FL( fprintf( stderr, "------RUN VIRTUAL MACHINE: exit: ret == F_BEG\n" ) );
	  return F_BEG;
	}
	else if( ret == F_CATCH ) {
	  D_FL( fprintf( stderr, "------RUN VIRTUAL MACHINE: exit: ret == F_CATCH\n" ) );
	  return F_RETRY;
	}
	reverse_method = T_FWDS_INTO;
      } else if( reverse_method == T_BACK_OVER ) {
	ret = reverseVirtualMachine( meth, lcl, &sp, &npc, retval, mjbuf, tid,
				     debug, callback, meth->fullSourceFile, &opsobjs, lvtobjs, R_OVER, 
				     &cl, &seen_new_line, code, thr, false );
	assert( ret != F_END );
	if( ret == F_BEG ) {
	  D_FL( fprintf( stderr, "------RUN VIRTUAL MACHINE: exit: ret == F_BEG\n" ) );
	  return F_BEG;
	}
	else if( ret == F_CATCH ) {
	  D_FL( fprintf( stderr, "------RUN VIRTUAL MACHINE: exit: ret == F_CATCH\n" ) );
	  return F_RETRY_OVER;
	}
        assert( 0 );
      } else if( should_bpt ) {
	debug->breakpoint = 1;
	if( meth->haveBytecode && !meth->haveSource )
	  debug->bytecode = 1;
	copyBreakpoints( callback );
      }
    }

    /* Skip native methods in reverse direction */

    if( !valid_log && IS_BACKWARDS(reverse_method)
	&& !IS_RETRY( reverse_method ) ) {
      D_FL( fprintf( stderr, "------RUN VIRTUAL MACHINE: exit: unlogged backwards skip; method %s\n", meth->name->data ) );
      return F_BEG;
    }

    /* we don't want to allow two logged methods separated by
     * an unlogged one in the call chain: attempting to reverse
     * this yields a complete mess */
    if( !valid_log && ( reverse_method == T_FWDS_INTO || reverse_method == T_FWDS_OVER ) ) {
      reverse_method = T_DEEP;
      thr = -1;
    }
  }

  /* END JJC */

  /* Finally we get to actually execute the machine */
  for (;;) {
    register uintp pc = npc; 

    assert(npc < meth->c.bcode.codelen);
    mjbuf->pc = pc;
    mjbuf->opsobjs = opsobjs; /* JJC */
    mjbuf->opstop = sp; /* JJC */
    npc = pc + insnLen[code[pc]];

    /* START JJC (if appropriate, wait for command and act on it) */
    if( valid_log ) {
      cline = meth->haveSource ? currentLine( pc, meth ) : currentPc( pc );

      if( cl != cline || debug->bytecode )
	seen_new_line = true;

      if( seen_new_line ) {

	if( debug->breakpoint ) 
	  IS_BREAKPOINT( debug->bytecode ? pc: cline, debug->bytecode );
	else
	  is_breakpoint = 1;
	if( is_breakpoint && ( reverse_method != T_FWDS_OVER ) 
	    && !IS_RETRY( reverse_method )
	    && ( meth->deepLog == false ) ) {
	  if( meth->haveSource )
	    CCreportNewLine( (jint)cline, (jint)(meth->fileOffset), (jint)( callback->user_number ),
			     (!( debug->bytecode && meth->haveBytecode) ) );
	  if( meth->haveBytecode ) 
	    CCreportNewPc( (jint)pc, 
			   utf8Const2Java( meth->name ),
			   utf8Const2Java( meth->parsed_sig->signature ),
			   (jint)(meth->bytecodeFileOffset), 
			   (jint)( callback->user_number ),
			   (!( !(debug->bytecode) && meth->haveSource)) );

	  D_FL( fprintf( stderr, "RUN VIRTUAL MACHINE: stack depth is %d, method is %s\n", sp - osp, meth->name->data ) );

	  debug->breakpoint = 0;
	  debug->bytecode = meth->haveSource ? 0 : 1;
	  running_over = false;

	  if( debug->test_on_next_stop ) {
	    if( debug->cbuffer_p_test != debug->cbuffer_p ||
		debug->cbuffer_v_test != debug->cbuffer_v ||
		debug->cbufferpc_p_test != debug->cbufferpc_p ||
		debug->cbufferpc_v_test != debug->cbufferpc_v ) {
	      fprintf( stderr, "-----------------------------------------\n" ); 
	      fprintf( stderr, "Ptrs: cbp %d cpv %d pcp %d pcv %d\n",
		       debug->cbuffer_p - debug->cbuffer_s,
		       debug->cbuffer_v - debug->cbuffer_s,
		       debug->cbufferpc_p - debug->cbufferpc_s,
		       debug->cbufferpc_v - debug->cbufferpc_s );
	      fprintf( stderr, "Expd: cbp %d cpv %d pcp %d pcv %d\n",
		       debug->cbuffer_p_test - debug->cbuffer_s,
		       debug->cbuffer_v_test - debug->cbuffer_s,
		       debug->cbufferpc_p_test - debug->cbufferpc_s,
		       debug->cbufferpc_v_test - debug->cbufferpc_s );
	      fprintf( stderr, "-----------------------------------------\n" ); 
	    }
	  }            

	loop: debug->test_on_next_stop = 0;
	  CCwritePrompt( (jint)( callback->user_number ) );
	  switch( (int)( CCgetCommand( (jint)( callback->user_number ) ) ) ) {
	  default:
	    fprintf( stderr, "Invalid debugger instruction\n" );
	    break;
	  case C_BYTECODE_STEPOVER:
	    running_over = true;          
	    /* fall thru */
	  case C_BYTECODE_STEP:
	    debug->bytecode = 1;
	    cl = cline;
	    seen_new_line = true;
	    break;
	  case C_STEPOVER:
	    running_over = true;
	    /* fall thru */
	  case C_STEP: 
	    cl = cline;
	    seen_new_line = false;
	    break; 
	  case C_BYTECODE_BACKBPT:
	    debug->bytecode = 1;
	    /* fall thru */
	  case C_BACKBPT: {
	    copyBreakpoints( callback );
	    debug->breakpoint = 1;
	    if( !doSimulation( meth, debug, callback, R_INTO, mjbuf ) ) {
	      CCcantBack( );
	      debug->breakpoint = 0;
	      goto loop;
	    } 
	    debug->breakpoint = 1;
	    ret = reverseVirtualMachine( meth, lcl, &sp, &npc, retval, mjbuf, tid,
					 debug, callback, meth->fullSourceFile, &opsobjs, lvtobjs, R_INTO, 
					 &cl, &seen_new_line, code, thr, false );
	    if( ret == F_BEG ) {
	      D_FL( fprintf( stderr, "------RUN VIRTUAL MACHINE: exit: ret == F_BEG\n" ) );
	      return F_BEG;
	    }
	    else if( ret == F_CATCH ) {
	      D_FL( fprintf( stderr, "------RUN VIRTUAL MACHINE: exit: ret == F_CATCH\n" ) );
	      return F_RETRY;
	    }
	    continue;
	  }
	  case C_BYTECODE_STEPBACKOVER:
	    debug->bytecode = 1;
	    /* fall thru */  
	  case C_STEPBACKOVER: {
	    if( !doSimulation( meth, debug, callback, R_OVERFIRST, mjbuf ) ) {
	      CCcantBack( );
	      goto loop;
	    }
	    ret = reverseVirtualMachine( meth, lcl, &sp, &npc, retval, mjbuf, tid,
					 debug, callback, meth->fullSourceFile, &opsobjs, lvtobjs, R_OVERFIRST, 
					 &cl, &seen_new_line, code, thr, false );
	    if( ret == F_BEG ) {
	      D_FL( fprintf( stderr, "------RUN VIRTUAL MACHINE: exit: ret == F_BEG\n" ) );
	      return F_BEG;
	    }
	    else if( ret == F_CATCH ) {
	      D_FL( fprintf( stderr, "------RUN VIRTUAL MACHINE: exit: ret == F_CATCH\n" ) );
	      return F_RETRY_OVER_FIRST;
	    }
	    continue;
	  }
	  case C_BYTECODE_STEPBACK:
	    debug->bytecode = 1;
	    /* fall thru */
	  case C_STEPBACK: {
	    if( !doSimulation( meth, debug, callback, R_INTO, mjbuf ) ) {
	      CCcantBack( );
	      goto loop; 
	    }
	    ret = reverseVirtualMachine( meth, lcl, &sp, &npc, retval, mjbuf, tid,
					 debug, callback, meth->fullSourceFile, &opsobjs, lvtobjs, R_INTO, 
					 &cl, &seen_new_line, code, thr, false );
	    if( ret == F_BEG ) {
	      D_FL( fprintf( stderr, "------RUN VIRTUAL MACHINE: exit: ret == F_BEG\n" ) );
	      return F_BEG;
	    }
	    else if( ret == F_CATCH ) {
	      D_FL( fprintf( stderr, "------RUN VIRTUAL MACHINE: exit: ret == F_CATCH\n" ) );
	      return F_RETRY;
	    }
	    continue;
	  }
	  case C_GETWATCH:
	    debug->no_file_log = 1;
	    getwatch( pc, meth, lcl, callback, thr, mjbuf );
	    debug->no_file_log = 0;
	    goto loop;
	  case C_RESET_BUFFERS:
	    resetBuffers( debug, thr );
	    goto loop;
	  case C_BYTECODE_BREAKPOINT:
	    debug->bytecode = 1;
	    /* fall thru */
	  case C_BREAKPOINT:
	    copyBreakpoints( callback );
	    debug->breakpoint = 1;
	    cl = cline;
	    seen_new_line = false;
	  }
	}
      }
    }
		
    /* END JJC */
		
    /* START JJC (record data to be lost in next instruction) */

    if( valid_log ) {
      wcbuff_increment( debug );
      if( defaultReverse[ code[ pc ] ] )
	savestack( pops[ code[ pc ] ] );
      wcbuffpc( pc, debug );		  
    }	
       
    /* END JJC */
    switch ( code[ pc ] ) {
    default:
      fprintf(stderr, "Unknown bytecode %d\n", code[pc]);
      throwException(NEW_LANG_EXCEPTION(VerifyError));
      break;
#include "kaffe.def"
    }
  }
 end:   /* START JJC (deal with method exit) */
  if( valid_log ) { 	  
    D_FL( fprintf( stderr, "------RUN VIRTUAL MACHINE: exit: reached end of %s\n", meth->name->data ) );
    wcbuffpc( END_OF_METHOD, debug );

    if( is_main )
      reportThreadDeath( callback, debug, thr );
  } 
  /* END JJC */
  return F_END;
}

/*
 * say what engine we're using
 */
char* 
getEngine()
{
  return "kaffe.intr";
}
